(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = global || self, factory(global.HE = {}));
}(this, function (exports) { 'use strict';

    var _null = null,
        _undefined = undefined,
        _Infinity = Infinity;
    var _window = window,
        _document = _window.document,
        _Object = Object,
        _assign = _Object.assign,
        _Map = Map,
        _Set = Set,
        _Math = Math,
        _sqrt = _Math.sqrt,
        _pow = _Math.pow,
        _sin = _Math.sin,
        _cos = _Math.cos,
        _PI = _Math.PI,
        _max = _Math.max,
        _min = _Math.min,
        _abs = _Math.abs,
        _sign = _Math.sign,
        _now = Date.now,
        _setTimeout = setTimeout,
        _clearTimeout = clearTimeout;

    var Utils = {
      Const: {
        EMPTY_OBJECT: _Object.create(_null),
        DOUBLE_PI: _PI * 2,
        HALF_PI: _PI / 2,
        TRANSPARENT: 'rgba(0,0,0,0)',
        IS_TOUCH_MODE: navigator.maxTouchPoints > 0 || /iOS|iPhone|iPod|iPad/.test(navigator.userAgent)
      },
      removeIndex: function removeIndex(array, index) {
        if (!index) {
          array.shift();
        } else {
          var end = array.length - 1;

          if (index < end) {
            for (; index < end; index++) {
              array[index] = array[index + 1];
            }

            array.length--;
          } else {
            array.pop();
          }
        }
      },
      insert: function insert(array, index, item) {
        if (!index) {
          array.unshift(item);
        } else if (index < array.length) {
          var end = array.length++ - 1;

          for (var i = end; i >= index; i--) {
            array[i + 1] = array[i];
          }

          array[index] = item;
        } else {
          array.push(item);
        }
      },
      mix: function mix(a, b, k) {
        if (a === _Infinity) {
          if (b === a) {
            return a;
          } else if (b === -a) {
            return NaN;
          }
        } else if (a === -_Infinity) {
          if (b === a) {
            return a;
          } else if (b === -a) {
            return NaN;
          }
        }

        return a + (b - a) * k;
      },
      random: function random(min, max) {
        return Utils.mix(min, max, _Math.random());
      },
      pick: function pick(options) {
        return options[_Math.floor(options.length * _Math.random())];
      },
      rad2deg: function rad2deg(rad) {
        return rad / _PI * 180;
      },
      deg2rad: function deg2rad(deg) {
        return deg / 180 * _PI;
      },
      quadraticSum: function quadraticSum(a, b) {
        return a * a + b * b;
      },
      distance: function distance(x1, y1, x2, y2) {
        return _sqrt(Utils.quadraticSum(x2 - x1, y2 - y1));
      },
      debounce: _assign(function debounce(callback, delay) {
        var timer;

        var wrapper = function debounceWrapper() {
          if (timer !== _undefined) {
            _clearTimeout(timer);
          }

          timer = _setTimeout.apply(_window, [callback, wrapper.delay].concat(arguments));
        };

        wrapper.delay = delay || Utils.debounce.defaultDelay;
        return wrapper;
      }, {
        defaultDelay: 100
      }),
      throttle: _assign(function throttle(callback, threshold) {
        var wrapper = function throttleWrapper() {
          var currentTime = _now();

          if (currentTime - wrapper.lastCallTime >= wrapper.threshold) {
            wrapper.lastCallTime = currentTime;
            return callback.apply(this, arguments);
          }
        };

        wrapper.lastCallTime = 0;
        wrapper.threshold = threshold || Utils.throttle.defaultThreshold;
        return wrapper;
      }, {
        defaultThreshold: 100
      }),
      cache: function cache(fn) {
        var map = new _Map();

        var wrapper = function cacheWrapper() {
          for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
            args[_key] = arguments[_key];
          }

          var parameters = args.join();

          if (map.has(parameters)) {
            return map.get(parameters);
          } else {
            var result = fn.apply(this, args);
            map.set(parameters, result);
            return result;
          }
        };

        wrapper.cache = map;
        return wrapper;
      }
    };

    var EventEmitter =
    /*#__PURE__*/
    function () {
      function EventEmitter() {
        this._listenerMap = new _Map();
        this.on = this.on.bind(this);
        this.off = this.off.bind(this);
        this.once = this.once.bind(this);
        this.emit = this.emit.bind(this);
        this.clearEvent = this.clearEvent.bind(this);
        this.clearEvents = this.clearEvents.bind(this);
      }

      var _proto = EventEmitter.prototype;

      _proto.on = function on(event, listener, once) {
        var _listenerMap = this._listenerMap;

        if (_listenerMap.has(event)) {
          _listenerMap.get(event).push([listener, !once]);
        } else {
          _listenerMap.set(event, [[listener, !once]]);
        }

        return this;
      };

      _proto.off = function off(event, listener, once) {
        once = !once;
        var _listenerMap = this._listenerMap;

        if (_listenerMap.has(event)) {
          var records = _listenerMap.get(event),
              index = records.findIndex(function (record) {
            return record[0] === listener && record[1] === once;
          });

          if (!~index) {
            Utils.removeIndex(records, index);
          }
        }

        return this;
      };

      _proto.once = function once(event, listener) {
        return this.on(event, listener, true);
      };

      _proto.emit = function emit(event) {
        var _this = this;

        for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
          args[_key - 1] = arguments[_key];
        }

        var _listenerMap = this._listenerMap;

        if (_listenerMap.has(event)) {
          var records = _listenerMap.get(event);

          if (records.length) {
            _listenerMap.set(event, records.filter(function (record) {
              record[0].apply(_this, args);
              return record[1];
            }));

            return true;
          }
        }

        return false;
      };

      _proto.clearEvent = function clearEvent(event) {
        this._listenerMap["delete"](event);

        return this;
      };

      _proto.clearEvents = function clearEvents() {
        this._listenerMap.clear();

        return this;
      };

      return EventEmitter;
    }();

    function _defineProperties(target, props) {
      for (var i = 0; i < props.length; i++) {
        var descriptor = props[i];
        descriptor.enumerable = descriptor.enumerable || false;
        descriptor.configurable = true;
        if ("value" in descriptor) descriptor.writable = true;
        Object.defineProperty(target, descriptor.key, descriptor);
      }
    }

    function _createClass(Constructor, protoProps, staticProps) {
      if (protoProps) _defineProperties(Constructor.prototype, protoProps);
      if (staticProps) _defineProperties(Constructor, staticProps);
      return Constructor;
    }

    function _inheritsLoose(subClass, superClass) {
      subClass.prototype = Object.create(superClass.prototype);
      subClass.prototype.constructor = subClass;
      subClass.__proto__ = superClass;
    }

    function _assertThisInitialized(self) {
      if (self === void 0) {
        throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
      }

      return self;
    }

    var Pool =
    /*#__PURE__*/
    function () {
      function Pool(factory, options) {
        this.factory = factory;
        this.options = options;
        this._items = new Array();
      }

      var _proto = Pool.prototype;

      _proto.get = function get() {
        var _items = this._items;

        if (_items.length) {
          return _items.shift();
        } else {
          return new this.factory(this.options);
        }
      };

      _proto.add = function add(item) {
        this._items.push(item);

        return this;
      };

      _createClass(Pool, [{
        key: "size",
        get: function get() {
          return this._items.length;
        },
        set: function set(value) {
          var _items = this._items,
              currentSize = _items.length;
          _items.length = value;

          if (currentSize < value) {
            for (var i = currentSize; i < value; i++) {
              _items[i] = new this.factory(this.options);
            }
          }
        }
      }]);

      return Pool;
    }();

    var interpolate = function interpolate(p1, p2, t) {
      var _t = 1 - t;

      return 3 * p1 * _t * _t * t + 3 * p2 * t * t * _t + t * t * t;
    };

    var cubic = function cubic(x1, y1, x2, y2) {
      return function (x) {
        if (x <= 0) {
          return 0;
        } else if (x >= 1) {
          return 1;
        }

        var accuracy = cubic.accuracy;
        var l = 0,
            r = 1,
            t = .5,
            cur;

        while (_abs((cur = interpolate(x1, x2, t)) - x) > accuracy) {
          if (cur > x) {
            r = t;
          } else {
            l = t;
          }

          t = (l + r) / 2;
        }

        return interpolate(y1, y2, t);
      };
    };

    cubic.accuracy = .001;
    var Timing = {
      linear: function linear(x) {
        return x;
      },
      cubic: cubic,
      ease: cubic(.25, .2, .25, 1),
      easeIn: cubic(.42, 0, 1, 1),
      easeOut: cubic(0, 0, .58, 1),
      easeInOut: cubic(.42, 0, 0.25, 1),
      steps: function steps(stepCount, start) {
        var step = 1 / stepCount;
        return function (x) {
          var t = _Math.floor(x / step);

          if (start && t < stepCount) {
            t += 1;
          }

          return t * step;
        };
      }
    };

    var Transition =
    /*#__PURE__*/
    function (_EventEmitter) {
      _inheritsLoose(Transition, _EventEmitter);

      function Transition(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _EventEmitter.call(this) || this;
        _this._delay = 0;

        _assign(_assertThisInitialized(_this), Transition.defaults, options);

        if (_this.active) {
          _this._startTime = _now();
        }

        var _assertThisInitialize = _assertThisInitialized(_this),
            target = _assertThisInitialize.target,
            key = _assertThisInitialize.key;

        if (options.from === _undefined && target) {
          _this.from = target[key];
        }

        if (options.to === _undefined && target) {
          _this.to = target[key];
        }

        return _this;
      }

      var _proto = Transition.prototype;

      _proto.start = function start() {
        this._startTime = _now();
        this.active = true;
        this._delay = 0;
        this.emit('start');
      };

      _proto.update = function update() {
        if (!this.active) {
          return;
        }

        var from = this.from,
            to = this.to;
        var x = (_now() - this._startTime - this.delay - this._delay) / this.duration,
            value;

        if (x >= 1) {
          value = to;
        } else if (x <= 0) {
          value = from;
        } else {
          value = Utils.mix(from, to, this.timing(x));
        }

        var target = this.target;

        if (target) {
          target[this.key] = value;
        }

        this.emit('update', value);

        if (x >= 1) {
          this.active = false;
          this.emit('end');

          if (this.loop) {
            this.start();
          }
        }
      };

      _proto.pause = function pause() {
        this.active = false;
        this._pauseTime = _now();
      };

      _proto.resume = function resume() {
        this._delay += _now() - this._pauseTime;
        this.active = true;
      };

      _proto.finish = function finish() {
        this.emit('update', this.to);
        this.active = false;
        this.emit('end');
      };

      return Transition;
    }(EventEmitter);
    Transition.defaults = {
      active: true,
      from: 0,
      to: 1,
      duration: 1000,
      delay: 0,
      loop: false,
      timing: Timing.linear,
      target: _null
    };

    var Vector =
    /*#__PURE__*/
    function () {
      Vector.of = function of(x, y) {
        return new Vector(x, y);
      };

      Vector.from = function from(vectorLike) {
        return Vector.of(vectorLike.x, vectorLike.y);
      };

      Vector.plus = function plus(vectors) {
        return vectors.reduce(function (ans, cur) {
          return ans.plusVector(cur);
        }, new Vector());
      };

      Vector.mean = function mean(vectors) {
        return Vector.plus(vectors).shrink(vectors.length);
      };

      Vector.dot = function dot(vector1, vector2) {
        return vector1.x * vector2.x + vector1.y * vector2.y;
      };

      Vector.cross = function cross(vector1, vector2) {
        return vector1.x * vector2.y - vector1.y * vector2.x;
      };

      Vector.minus = function minus(vector1, vector2) {
        return Vector.of(vector1.x - vector2.x, vector1.y - vector2.y);
      };

      Vector.project = function project(sourceVector, directionVector) {
        var dirVecMod = directionVector.getNorm();
        return dirVecMod && Vector.dot(sourceVector, directionVector) / dirVecMod;
      };

      Vector.projectVector = function projectVector(sourceVector, directionVector) {
        return directionVector.clone().setNorm(Vector.project(sourceVector, directionVector));
      };

      Vector.distance = function distance(vector1, vector2) {
        return Utils.distance(vector1.x, vector1.y, vector2.x, vector2.y);
      };

      Vector.distribute = function distribute(vector0, vector1, vector2, k1, k2, scale) {
        var sum = _abs(k1) + _abs(k2);

        if (scale !== _undefined) {
          sum /= scale;
        }

        k1 /= sum;
        k2 /= sum;
        vector1.x += vector0.x * k1;
        vector1.y += vector0.y * k1;
        vector2.x += vector0.x * k2;
        vector2.y += vector0.y * k2;
      };

      Vector.random = function random(startAngle, endAngle) {
        if (endAngle === void 0) {
          endAngle = Utils.Const.DOUBLE_PI;
        }

        var angle = Utils.random(startAngle || 0, endAngle);
        return Vector.of(_cos(angle), _sin(angle));
      };

      function Vector(x, y) {
        this.x = x || 0;
        this.y = y || 0;
      }

      var _proto = Vector.prototype;

      _proto.clone = function clone() {
        return Vector.of(this.x, this.y);
      };

      _proto.set = function set(x, y) {
        this.x = x;
        this.y = y;
        return this;
      };

      _proto.setVector = function setVector(vector, scale) {
        if (scale !== _undefined) {
          this.x = vector.x * scale;
          this.y = vector.y * scale;
        } else {
          this.x = vector.x;
          this.y = vector.y;
        }

        return this;
      };

      _proto.reset = function reset() {
        this.x = this.y = 0;
        return this;
      };

      _proto.getNorm = function getNorm() {
        return _sqrt(_pow(this.x, 2) + _pow(this.y, 2));
      };

      _proto.setNorm = function setNorm(norm) {
        var currentNorm = this.getNorm();

        if (currentNorm) {
          var scale = norm / currentNorm;
          this.x *= scale;
          this.y *= scale;
        }

        return this;
      };

      _proto.plus = function plus(dx, dy) {
        this.x += dx;
        this.y += dy;
        return this;
      };

      _proto.plusVector = function plusVector(vector, scale) {
        if (scale !== _undefined) {
          this.x += vector.x * scale;
          this.y += vector.y * scale;
        } else {
          this.x += vector.x;
          this.y += vector.y;
        }

        return this;
      };

      _proto.minus = function minus(dx, dy) {
        this.x -= dx;
        this.y -= dy;
        return this;
      };

      _proto.minusVector = function minusVector(vector, scale) {
        if (scale !== _undefined) {
          this.x -= vector.x * scale;
          this.y -= vector.y * scale;
        } else {
          this.x -= vector.x;
          this.y -= vector.y;
        }

        return this;
      };

      _proto.scale = function scale(scaleX, scaleY, origin) {
        if (scaleY === void 0) {
          scaleY = scaleX;
        }

        if (origin) {
          this.x = origin.x + (this.x - origin.x) * scaleX;
          this.y = origin.y + (this.y - origin.y) * scaleY;
        } else {
          this.x *= scaleX;
          this.y *= scaleY;
        }

        return this;
      };

      _proto.shrink = function shrink(shrinkX, shrinkY, origin) {
        if (shrinkY === void 0) {
          shrinkY = shrinkX;
        }

        if (origin) {
          this.x = origin.x + (this.x - origin.x) / shrinkX;
          this.y = origin.y + (this.y - origin.y) / shrinkY;
        } else {
          this.x /= shrinkX;
          this.y /= shrinkY;
        }

        return this;
      };

      _proto.grow = function grow(delta) {
        var norm = this.getNorm();

        if (norm) {
          var scale = (norm + delta) / norm;
          this.x *= scale;
          this.y *= scale;
        }

        return this;
      };

      _proto.rotate = function rotate(radian, origin) {
        var cos = _cos(radian),
            sin = _sin(radian);

        if (origin) {
          var dx = this.x - origin.x,
              dy = this.y - origin.y;
          this.x = origin.x + dx * cos - dy * sin;
          this.y = origin.y + dx * sin + dy * cos;
        } else {
          var x = this.x,
              y = this.y;
          this.x = x * cos - y * sin;
          this.y = x * sin + y * cos;
        }

        return this;
      };

      _proto.getAngle = function getAngle() {
        var x = this.x;
        return _Math.atan(this.y / x) + (x > 0 ? 0 : _PI);
      };

      _proto.turn = function turn(clockwise) {
        var x = this.x,
            y = this.y;

        if (clockwise) {
          this.x = y;
          this.y = -x;
        } else {
          this.x = -y;
          this.y = x;
        }

        return this;
      };

      _proto.reverse = function reverse() {
        this.x = -this.x;
        this.y = -this.y;
        return this;
      };

      _proto.normalize = function normalize() {
        var norm = this.getNorm();
        this.x /= norm;
        this.y /= norm;
        return this;
      };

      _proto.isZero = function isZero() {
        return this.x === 0 && this.y === 0;
      };

      _proto.toString = function toString(fractionDigits) {
        return this.x.toFixed(fractionDigits) + ',' + this.y.toFixed(fractionDigits);
      };

      return Vector;
    }();

    var Bounds =
    /*#__PURE__*/
    function () {
      function Bounds() {
        this.top = 0;
        this.right = 0;
        this.bottom = 0;
        this.left = 0;
      }

      var _proto = Bounds.prototype;

      _proto.move = function move(deltaX, deltaY) {
        this.left += deltaX;
        this.right += deltaX;
        this.top += deltaY;
        this.bottom += deltaY;
      };

      _proto.moveVector = function moveVector(vector, scale) {
        if (scale) {
          this.left += vector.x * scale;
          this.right += vector.x * scale;
          this.top += vector.y * scale;
          this.bottom += vector.y * scale;
        } else {
          this.left += vector.x;
          this.right += vector.x;
          this.top += vector.y;
          this.bottom += vector.y;
        }
      };

      _proto.grow = function grow(delta) {
        this.left -= delta;
        this.right += delta;
        this.top -= delta;
        this.bottom += delta;
      };

      _proto.overlaps = function overlaps(bounds) {
        return !(bounds.right < this.left || bounds.left > this.right || bounds.bottom < this.top || bounds.top > this.bottom);
      };

      _proto.updateVertices = function updateVertices(vertices) {
        var _this = this;

        vertices.forEach(function (_ref, i) {
          var x = _ref.x,
              y = _ref.y;

          if (i > 0) {
            if (x < _this.left) {
              _this.left = x;
            } else if (x > _this.right) {
              _this.right = x;
            }

            if (y < _this.top) {
              _this.top = y;
            } else if (y > _this.bottom) {
              _this.bottom = y;
            }
          } else {
            _this.left = _this.right = x;
            _this.top = _this.bottom = y;
          }
        });
      };

      _proto.updateBounds = function updateBounds(boundsArray) {
        var _this2 = this;

        boundsArray.forEach(function (_ref2, i) {
          var left = _ref2.left,
              right = _ref2.right,
              top = _ref2.top,
              bottom = _ref2.bottom;

          if (i > 0) {
            if (left < _this2.left) {
              _this2.left = left;
            }

            if (right > _this2.right) {
              _this2.right = right;
            }

            if (top < _this2.top) {
              _this2.top = top;
            }

            if (bottom > _this2.bottom) {
              _this2.bottom = bottom;
            }
          } else {
            _this2.left = left;
            _this2.right = right;
            _this2.top = top;
            _this2.bottom = bottom;
          }
        });
      };

      _createClass(Bounds, [{
        key: "width",
        get: function get() {
          return this.right - this.left;
        }
      }, {
        key: "height",
        get: function get() {
          return this.bottom - this.top;
        }
      }]);

      return Bounds;
    }();

    var Vertices = {
      fromArray: function fromArray(array) {
        var vertices = [];

        for (var i = 0; i < array.length; i += 2) {
          vertices.push(Vector.of(array[i], array[i + 1]));
        }

        return vertices;
      },
      findClosest: function findClosest(target, vertices) {
        var min = _Infinity,
            closest;
        vertices.forEach(function (vertex) {
          var current = Utils.quadraticSum(target.x - vertex.x, target.y - vertex.y);

          if (current < min) {
            min = current;
            closest = vertex;
          }
        });
        return closest;
      },
      createPolygon: Utils.cache(function (edges, radius, rotation) {
        if (rotation === void 0) {
          rotation = -Utils.Const.HALF_PI;
        }

        var angle = Utils.Const.DOUBLE_PI / edges,
            results = [];

        for (var i = 0; i < edges; i++) {
          results.push(Vector.of(_cos(rotation), _sin(rotation)).setNorm(radius));
          rotation += angle;
        }

        return results;
      }),
      createStar: Utils.cache(function (angles, innerRadius, outerRadius, rotation) {
        if (rotation === void 0) {
          rotation = -Utils.Const.HALF_PI;
        }

        var angle = _PI / angles,
            results = [];

        for (var i = 0; i < angles; i++) {
          results.push(Vector.of(_cos(rotation), _sin(rotation)).setNorm(innerRadius));
          rotation += angle;
          results.push(Vector.of(_cos(rotation), _sin(rotation)).setNorm(outerRadius));
          rotation += angle;
        }

        return results;
      }),
      createRectangle: Utils.cache(function (width, height, rotation) {
        var x0 = width / 2,
            y0 = height / 2;

        if (rotation) {
          var cos = _cos(rotation),
              sin = _sin(rotation),
              x1 = x0 * cos - y0 * sin,
              y1 = x0 * sin + y0 * sin,
              x2 = -x0 * cos - y0 * sin,
              y2 = -x0 * sin + y0 * sin;

          return Vertices.fromArray([x1, y1, x2, y2, -x1, -y1, -x2, -y2]);
        } else {
          return Vertices.fromArray([x0, y0, -x0, y0, -x0, -y0, x0, -y0]);
        }
      })
    };

    var Category = {
      MAX_COUNT: 32,
      FULL_MASK: 4294967295,
      registry: [],
      count: 0,
      next: function next() {
        if (Category.count > Category.MAX_COUNT) {
          throw "Max category count exceeded";
        }

        return 1 << Category.count++;
      },
      "for": function _for(tag) {
        var record = Category.registry.find(function (record) {
          return record[0] === tag;
        });

        if (!record) {
          var category = Category.next();
          Category.registry.push([tag, category]);
          return category;
        } else {
          return record[1];
        }
      },
      tagFor: function tagFor(category) {
        var result = Category.registry.find(function (record) {
          return record[1] === category;
        });
        return result && result[0];
      }
    };

    var Body =
    /*#__PURE__*/
    function (_EventEmitter) {
      _inheritsLoose(Body, _EventEmitter);

      function Body(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _EventEmitter.call(this) || this;
        _this.tag = '';
        _this.bounds = new Bounds();
        _this.area = 0;
        _this.normals = [];
        _this.scaleX = 1;
        _this.scaleY = 1;
        _this.rotation = 0;
        _this.mass = 0;
        _this.speed = 0;
        _this.contact = new _Set();

        _assign(_assertThisInitialized(_this), Body.defaults, options);

        if (!_this.position) {
          _this.position = new Vector();
        }

        if (!_this.velocity) {
          _this.velocity = new Vector();
        }

        if (!_this.acceleration) {
          _this.acceleration = new Vector();
        }

        if (_this.category) {
          _this.tag = Category.tagFor(_this.category) || '';
        } else if (_this.tag) {
          _this.category = Category["for"](_this.tag);
        }

        var _assertThisInitialize = _assertThisInitialized(_this),
            scaleX = _assertThisInitialize.scaleX,
            scaleY = _assertThisInitialize.scaleY,
            rotation = _assertThisInitialize.rotation;

        if (scaleX !== _undefined || scaleY !== _undefined) {
          _this.scale(scaleX, scaleY);
        }

        if (rotation) {
          _this.rotate(rotation);
        }

        if (_this.draggable) {
          _this.interactive = true;
        }

        if (_this.friction > _this.staticFriction) {
          _this.friction = _this.staticFriction;
        }

        return _this;
      }

      var _proto = Body.prototype;

      _proto.setDensity = function setDensity(density) {
        this.density = density;

        if (this.active) {
          this.mass = this.area * density;
        }

        return this;
      };

      _proto.setMass = function setMass(mass) {
        if (this.active) {
          this.mass = mass;
        }

        this.density = mass / this.area;
        return this;
      };

      _proto._setArea = function _setArea(area) {
        this.area = area;
        this.mass = this.density * area;
      };

      _proto.activate = function activate(active) {
        if (active === void 0) {
          active = true;
        }

        this.mass = (this.active = active) ? this.area * this.density : _Infinity;
        return this;
      };

      _proto.setScale = function setScale(scaleX, scaleY) {
        if (scaleY === void 0) {
          scaleY = scaleX;
        }

        var deltaScaleX = scaleX / this.scaleX,
            deltaScaleY = scaleY / this.scaleY,
            scale = deltaScaleX * deltaScaleY;
        this.area *= scale;
        this.mass *= scale;

        if (this._scale) {
          this._scale(deltaScaleX, deltaScaleY);
        }

        this.scaleX = scaleX;
        this.scaleY = scaleY;
        return this;
      };

      _proto.scale = function scale(scaleX, scaleY, origin) {
        if (scaleY === void 0) {
          scaleY = scaleX;
        }

        var scale = scaleX * scaleY;
        this.area *= scale;
        this.mass *= scale;

        if (this._scale) {
          this._scale(scaleX, scaleY, origin);
        }

        this.scaleX *= scaleX;
        this.scaleY *= scaleY;

        if (origin) {
          this.position.scale(scaleX, scaleY, origin);
        }

        return this;
      };

      _proto.shrink = function shrink(shrinkX, shrinkY, origin) {
        if (shrinkY === void 0) {
          shrinkY = shrinkX;
        }

        var shrink = shrinkX * shrinkY;
        this.area /= shrink;
        this.mass /= shrink;

        if (this._scale) {
          this._scale(1 / shrinkX, 1 / shrinkY, origin);
        }

        this.scaleX /= shrinkX;
        this.scaleY /= shrinkY;

        if (origin) {
          this.position.shrink(shrinkX, shrinkY, origin);
        }

        return this;
      };

      _proto.setRotation = function setRotation(rotation, origin) {
        var deltaRotation = rotation - this.rotation;

        if (this._rotate) {
          this._rotate(deltaRotation, origin);
        }

        this.rotation = rotation;

        if (origin) {
          this.position.rotate(deltaRotation, origin);
        }

        return this;
      };

      _proto.rotate = function rotate(rotation, origin) {
        if (this._rotate) {
          this._rotate(rotation, origin);
        }

        this.rotation += rotation;

        if (origin) {
          this.position.rotate(rotation, origin);
        }

        return this;
      };

      _proto.applyForce = function applyForce(force, point) {
        var mass = this.mass;
        this.acceleration.plus(force.x / mass, force.y / mass); // TODO: update angular speed

        return this;
      };

      _proto.move = function move(deltaX, deltaY) {
        this.position.plus(deltaX, deltaY);
        this.bounds.move(deltaX, deltaY);
        return this;
      };

      _proto.moveVector = function moveVector(vector, scale) {
        this.position.plusVector(vector, scale);
        this.bounds.moveVector(vector, scale);
        return this;
      };

      _proto.moveTo = function moveTo(x, y) {
        var position = this.position,
            deltaX = x - position.x,
            deltaY = y - position.y;
        this.position.plus(deltaX, deltaY);
        this.bounds.move(deltaX, deltaY);
        return this;
      };

      _proto.moveToVector = function moveToVector(vector) {
        var deltaVector = Vector.minus(vector, this.position);
        this.position.plusVector(deltaVector);
        this.bounds.moveVector(deltaVector);
        return this;
      };

      _proto.update = function update(timeScale) {
        this.emit('willUpdate', timeScale);

        if (this.active) {
          var velocity = this.velocity,
              bounds = this.bounds,
              position = this.position,
              maxSpeed = this.maxSpeed;
          var speed = velocity.getNorm() / timeScale;

          if (speed > maxSpeed) {
            velocity.scale(maxSpeed / speed);
            speed = maxSpeed;
          }

          position.plusVector(velocity, timeScale);
          bounds.moveVector(velocity, timeScale);
          this.isStatic = (this.speed = speed) <= Body.maxStaticSpeed;
          var acceleration = this.acceleration,
              gravity = this.gravity,
              maxAngularSpeed = this.maxAngularSpeed * timeScale,
              airSpeedScale = 1 - this.airFriction,
              angularSpeed = (this.angularSpeed *= airSpeedScale) * timeScale;
          velocity.plusVector(acceleration, timeScale);

          if (gravity) {
            velocity.plusVector(gravity, timeScale);
          }

          velocity.scale(airSpeedScale);
          acceleration.reset();

          if (angularSpeed > maxAngularSpeed) {
            this.rotation += this.angularSpeed = maxAngularSpeed;
          } else {
            this.rotation += angularSpeed;
          }

          if (this.fixRotation) {
            this.rotation %= Utils.Const.DOUBLE_PI;
          }
        } else {
          this.isStatic = false;
        }

        this.emit('didUpdate', timeScale);
      };

      _createClass(Body, [{
        key: "x",
        get: function get() {
          return this.position.x;
        },
        set: function set(value) {
          var position = this.position;
          this.bounds.move(value - position.x, 0);
          position.x = value;
        }
      }, {
        key: "y",
        get: function get() {
          return this.position.y;
        },
        set: function set(value) {
          var position = this.position;
          this.bounds.move(0, value - position.y);
          position.y = value;
        }
      }]);

      return Body;
    }(EventEmitter);
    Body.normalPrecision = 3;
    Body.maxStaticSpeed = 1;
    Body.defaults = {
      category: 0,
      collisionFilter: Category.FULL_MASK,
      sensorFilter: 0,
      isCircle: false,
      active: false,
      interactive: false,
      draggable: false,
      defer: false,
      maxSpeed: 200,
      angularSpeed: 0,
      maxAngularSpeed: 100,
      gravity: Vector.of(0, 2),
      density: 1,
      stiffness: 1,
      slop: .1,
      elasticity: .3,
      friction: .4,
      staticFriction: .5,
      airFriction: 0,
      fixRotation: true,
      radius: 0
    };

    var Collision = {
      check: function check(bodies, checker) {
        var maxStaticSpeed = Body.maxStaticSpeed;
        Collision.find(bodies, checker).filter(function (collisionInfo) {
          var body1 = collisionInfo.body1,
              body2 = collisionInfo.body2,
              overlapVector = collisionInfo.overlapVector,
              v1 = body1.velocity,
              stiffness1 = body1.stiffness,
              slop1 = body1.slop,
              m1 = body1.mass,
              elasticity1 = body1.elasticity;
          overlapVector.reverse();
          body1.emit('collision', body2, collisionInfo);
          overlapVector.reverse();
          body2.emit('collision', body1, collisionInfo);

          if (body1.category & body2.sensorFilter || body1.sensorFilter & body2.category) {
            return false;
          }

          var v2 = body2.velocity,
              m2 = body2.mass,
              elasticity = _min(elasticity1, body2.elasticity) + 1,
              slop = slop1 + body2.slop,
              impulse = collisionInfo.overlap * (stiffness1 + body2.stiffness) / 2,
              impulseScale = impulse > slop ? (impulse - slop) / impulse : 0,
              edgeVector = collisionInfo.edgeVector,
              relativeVelocity = collisionInfo.relativeVelocity,
              isSeparating = Vector.dot(relativeVelocity, overlapVector) < 0,
              relativeNormalVelocity = Vector.projectVector(relativeVelocity, overlapVector);

          if (body1.active) {
            if (body2.active) {
              if (impulseScale) {
                v1.minusVector(overlapVector, impulseScale / 2);
                v2.plusVector(overlapVector, impulseScale / 2);
              }

              if (edgeVector && isSeparating) {
                Vector.distribute(relativeNormalVelocity, v1, v2, m2, -m1, elasticity);
              }
            } else {
              if (impulseScale) {
                v1.minusVector(overlapVector, impulseScale);
              }

              if (edgeVector && isSeparating) {
                v1.plusVector(relativeNormalVelocity, elasticity);
              }
            }
          } else {
            if (body2.active) {
              if (impulseScale) {
                v2.plusVector(overlapVector, impulseScale);
              }

              if (edgeVector && isSeparating) {
                v2.minusVector(relativeNormalVelocity, elasticity);
              }
            } else {
              return false;
            }
          }

          return edgeVector;
        }).forEach(function (collisionInfo) {
          // TODO: solve rotation
          var body1 = collisionInfo.body1,
              body2 = collisionInfo.body2;
          body1.contact.add(body2);
          body2.contact.add(body1);

          var v1 = body1.velocity,
              v2 = body2.velocity,
              friction = _min(body1.friction, body2.friction),
              staticFriction = _min(body1.staticFriction, body2.staticFriction);

          if (!staticFriction) {
            return;
          }

          var overlap = collisionInfo.overlap,
              edgeVector = collisionInfo.edgeVector,
              relativeVelocity = collisionInfo.relativeVelocity,
              relativeEdgeSpeed = Vector.project(relativeVelocity, edgeVector),
              absRelativeEdgeSpeed = _abs(relativeEdgeSpeed);

          if (body1.active) {
            if (body2.active) {
              if (absRelativeEdgeSpeed < maxStaticSpeed && overlap * staticFriction >= absRelativeEdgeSpeed || overlap * friction >= absRelativeEdgeSpeed) {
                var relativeEdgeVelocity = Vector.projectVector(relativeVelocity, edgeVector).shrink(2);
                v1.plusVector(relativeEdgeVelocity);
                v2.minusVector(relativeEdgeVelocity);
              } else if (friction) {
                Vector.distribute(edgeVector, v1, v2, body2.mass, -body1.mass, overlap * friction * _sign(relativeEdgeSpeed));
              }
            } else {
              if (absRelativeEdgeSpeed < maxStaticSpeed && overlap * staticFriction >= absRelativeEdgeSpeed || overlap * friction >= absRelativeEdgeSpeed) {
                v1.plusVector(edgeVector, relativeEdgeSpeed);
              } else if (friction) {
                v1.plusVector(edgeVector, overlap * friction * _sign(relativeEdgeSpeed));
              }
            }
          } else {
            if (absRelativeEdgeSpeed < maxStaticSpeed && overlap * staticFriction >= absRelativeEdgeSpeed || overlap * friction >= absRelativeEdgeSpeed) {
              v2.minusVector(edgeVector, relativeEdgeSpeed);
            } else if (friction) {
              v2.minusVector(edgeVector, overlap * friction * _sign(relativeEdgeSpeed));
            }
          }
        });
      },
      find: function find(bodies, checker) {
        var results = new Array(),
            bodiesCount = bodies.length;

        for (var i = 0; i < bodiesCount; i++) {
          var body1 = bodies[i],
              category1 = body1.category,
              velocity1 = body1.velocity;

          for (var j = i + 1; j < bodiesCount; j++) {
            var body2 = bodies[j];

            if (!(category1 & body2.collisionFilter && body1.collisionFilter & body2.category)) {
              continue;
            }

            var collisionInfo = checker(body1, body2);

            if (!collisionInfo) {
              continue;
            }

            var overlapVector = collisionInfo.overlapVector;
            results.push(_assign(collisionInfo, {
              body1: body1,
              body2: body2,
              edgeVector: overlapVector.isZero() ? _null : overlapVector.clone().turn().normalize(),
              relativeVelocity: Vector.minus(body2.velocity, velocity1)
            }));
          }
        }

        return results;
      },
      Checker: {
        AABB: function AABB(body1, body2) {
          var bounds1 = body1.bounds,
              bounds2 = body2.bounds;
          var delta = bounds1.right - bounds2.left,
              min = delta,
              x = delta,
              y = 0;

          if (delta < 0) {
            return _null;
          }

          delta = bounds1.left - bounds2.right;

          if (delta > 0) {
            return _null;
          } else if (-delta < min) {
            min = -delta;
            x = delta;
          }

          delta = bounds1.bottom - bounds2.top;

          if (delta < 0) {
            return _null;
          } else if (delta < min) {
            min = delta;
            x = 0;
            y = delta;
          }

          delta = bounds1.top - bounds2.bottom;

          if (delta > 0) {
            return _null;
          } else if (-delta < min) {
            min = -delta;
            x = 0;
            y = delta;
          }

          return {
            overlap: min,
            overlapVector: Vector.of(x, y)
          };
        },
        SAT: function SAT(body1, body2) {
          var position1 = body1.position;

          if (body2.isCircle && !body1.isCircle) {
            var result = Collision.Checker.SAT(body2, body1);

            if (result) {
              result.overlapVector.reverse();
            }

            return result;
          }

          var normals = body1.normals.concat(body2.normals, Vector.minus(position1, body2.getClosest(position1)).normalize());

          if (!body1.isCircle) {
            var position2 = body2.position;
            normals.push(Vector.minus(position2, body1.getClosest(position2)).normalize());
          }

          var minDirection = _null,
              minOverlap = _Infinity;

          for (var i = 0; i < normals.length; i++) {
            var direction = normals[i],
                projection1 = body1.project(direction),
                projection2 = body2.project(direction);

            if (projection1.min > projection2.max || projection1.max < projection2.min) {
              return _null;
            }

            var overlap1 = projection1.max - projection2.min,
                overlap2 = projection1.min - projection2.max;

            if (overlap1 < -overlap2) {
              if (overlap1 < minOverlap) {
                minOverlap = overlap1;
                minDirection = direction;
              }
            } else {
              if (-overlap2 < minOverlap) {
                minOverlap = -overlap2;
                minDirection = direction.reverse();
              }
            }
          }

          return minDirection && {
            overlap: minOverlap,
            overlapVector: minDirection.clone().scale(minOverlap)
          };
        },
        Distance: function Distance(body1, body2) {
          var offset = Vector.minus(body2.position, body1.position),
              delta = body1.radius + body2.radius - offset.getNorm();
          return delta < 0 ? _null : {
            overlap: delta,
            overlapVector: offset.setNorm(delta)
          };
        },
        Smart: function Smart(body1, body2) {
          if (body1.bounds.overlaps(body2.bounds)) {
            return body1.isCircle && body2.isCircle ? Collision.Checker.Distance(body1, body2) : Collision.Checker.SAT(body1, body2);
          } else {
            return _null;
          }
        }
      }
    };

    var Style = {
      Common: {
        defaults: {
          shadowColor: Utils.Const.TRANSPARENT,
          shadowBlur: 0,
          shadowOffsetX: 0,
          shadowOffsetY: 0,
          opacity: 1
        },
        apply: function apply(renderer, style) {
          var context = renderer.context,
              ratio = renderer.ratio;
          context.globalAlpha = style.opacity;
          context.shadowColor = style.shadowColor;
          context.shadowBlur = style.shadowBlur * ratio;
          context.shadowOffsetX = style.shadowOffsetX * ratio;
          context.shadowOffsetY = style.shadowOffsetY * ratio;
        }
      },
      Fill: {
        defaults: {
          fillStyle: _null
        },
        apply: function apply(renderer, style) {
          if (style.fillStyle) {
            var context = renderer.context;
            context.fillStyle = style.fillStyle;
          }
        }
      },
      Stroke: {
        defaults: {
          strokeStyle: '#000',
          lineWidth: 1,
          lineCap: 'butt',
          lineJoin: 'miter',
          miterLimit: 10,
          lineDash: _null,
          lineDashOffset: 0
        },
        apply: function apply(renderer, style) {
          if (!style.strokeStyle) {
            return;
          }

          var context = renderer.context;
          context.strokeStyle = style.strokeStyle;
          context.lineWidth = style.lineWidth;
          context.lineCap = style.lineCap;
          context.lineJoin = style.lineJoin;

          if (style.lineDash) {
            context.setLineDash(style.lineDash);
            context.lineDashOffset = style.lineDashOffset;
          }
        }
      }
    };

    var Constraint =
    /*#__PURE__*/
    function () {
      function Constraint(options) {
        _assign(this, Constraint.defaults, options);

        this.style = _assign({}, Constraint.defaultStyle, this.style);

        if (!this.originOffset) {
          this.originOffset = new Vector();
        }

        if (!this.targetOffset) {
          this.targetOffset = new Vector();
        }

        if (this.maxLength === _undefined) {
          var origin = this.origin,
              target = this.target;
          this.maxLength = origin && target ? Vector.distance(origin.position || origin, target.position) : 0;
        }

        if (this.minLength === _undefined) {
          this.minLength = 0;
        }
      }

      var _proto = Constraint.prototype;

      _proto.update = function update(timeScale) {
        var origin = this.origin,
            target = this.target;

        if (!target || !origin) {
          return;
        }

        var minLength = this.minLength,
            maxLength = this.maxLength,
            elasticity = this.elasticity,
            targetPosition = target.position,
            originPosition = origin.position || origin,
            offsetVector = Vector.minus(targetPosition, originPosition).minusVector(this.originOffset).plusVector(this.targetOffset),
            offset = offsetVector.getNorm(),
            delta = offset > maxLength ? offset - maxLength : offset < minLength ? offset - minLength : 0;

        if (!delta) {
          return;
        }

        offsetVector.setNorm(delta);
        var originIsActive = originPosition !== origin && origin.active,
            targetVelocity = target.velocity,
            stiffness = this.stiffness;

        if (originIsActive) {
          var originVelocity = origin.velocity;

          if (target.active) {
            var originMass = origin.mass,
                targetMass = target.mass,
                originScale = originMass / (originMass + targetMass),
                targetScale = 1 - originScale;
            origin.moveVector(offsetVector, originScale * stiffness);
            target.moveVector(offsetVector, -targetScale * stiffness);
            var relativeVelocity = Vector.minus(targetVelocity, originVelocity);

            if (Vector.dot(relativeVelocity, offsetVector)) {
              var bounceVelocity = Vector.projectVector(relativeVelocity, offsetVector).plusVector(offsetVector).scale(elasticity);
              originVelocity.plusVector(bounceVelocity, originScale);
              targetVelocity.minusVector(bounceVelocity, targetScale); // TODO: solve the rotations of origin & target here
            }
          } else {
            origin.moveVector(offsetVector);
            originVelocity.minusVector(Vector.projectVector(originVelocity, offsetVector), elasticity); // TODO: solve the rotation of origin here
          }
        } else {
          if (target.active || this.force) {
            target.move(-offsetVector.x, -offsetVector.y);
            targetVelocity.minusVector(Vector.projectVector(targetVelocity, offsetVector), elasticity); // TODO: solve the rotation of target here
          }
        }
      };

      _proto.render = function render(renderer) {
        var style = this.style,
            origin = this.origin,
            target = this.target;

        if (!style.strokeStyle || !origin || !target) {
          return;
        }

        var originOffset = this.originOffset,
            targetOffset = this.targetOffset,
            targetPosition = target.position,
            originPosition = origin.position || origin,
            context = renderer.context;
        Style.Common.apply(renderer, style);
        Style.Stroke.apply(renderer, style);
        context.beginPath();
        context.moveTo(originPosition.x + originOffset.x, originPosition.y + originOffset.y);
        context.lineTo(targetPosition.x + targetOffset.x, targetPosition.y + targetOffset.y);
        context.stroke();
      };

      _createClass(Constraint, [{
        key: "length",
        set: function set(length) {
          this.minLength = this.maxLength = length;
        }
      }]);

      return Constraint;
    }();
    Constraint.defaults = {
      origin: _null,
      target: _null,
      elasticity: 1,
      stiffness: 1,
      force: true,
      defer: true
    };
    Constraint.defaultStyle = _assign({}, Style.Common.defaults, Style.Stroke.defaults);

    var Sizing = {
      Full: function Full(innerWidth, innerHeight, outerWidth, outerHeight, margin) {
        var doubleMargin = margin * 2,
            width = outerWidth - doubleMargin,
            height = outerHeight - doubleMargin;
        return {
          width: width,
          height: height,
          styleWidth: width,
          styleHeight: height,
          left: margin,
          top: margin,
          scale: 1
        };
      },
      Center: function Center(innerWidth, innerHeight, outerWidth, outerHeight, margin) {
        return {
          width: innerWidth,
          height: innerHeight,
          styleWidth: innerWidth,
          styleHeight: innerHeight,
          left: (outerWidth - innerWidth) / 2,
          top: (outerHeight - innerHeight) / 2,
          scale: 1
        };
      },
      Fit: function Fit(innerWidth, innerHeight, outerWidth, outerHeight, margin) {
        var scale = outerWidth / outerHeight > innerWidth / innerHeight ? (outerHeight - margin * 2) / innerHeight : (outerWidth - margin * 2) / innerWidth,
            width = innerWidth * scale,
            height = innerHeight * scale;
        return {
          width: innerWidth,
          height: innerHeight,
          styleWidth: width,
          styleHeight: height,
          left: (outerWidth - width) / 2,
          top: (outerHeight - height) / 2,
          scale: scale
        };
      },
      FixedWidth: function FixedWidth(innerWidth, innerHeight, outerWidth, outerHeight, margin) {
        var doubleMargin = margin * 2,
            styleWidth = outerWidth - doubleMargin,
            styleHeight = outerHeight - doubleMargin,
            scale = styleWidth / innerWidth;
        return {
          width: innerWidth,
          height: styleHeight / scale,
          styleWidth: styleWidth,
          styleHeight: styleHeight,
          left: margin,
          top: margin,
          scale: scale
        };
      },
      FixedHeight: function FixedHeight(innerWidth, innerHeight, outerWidth, outerHeight, margin) {
        var doubleMargin = margin * 2,
            styleWidth = outerWidth - doubleMargin,
            styleHeight = outerHeight - doubleMargin,
            scale = styleHeight / innerHeight;
        return {
          width: styleWidth / scale,
          height: innerHeight,
          styleWidth: styleWidth,
          styleHeight: styleHeight,
          left: margin,
          top: margin,
          scale: scale
        };
      },
      Fixed: function Fixed(innerWidth, innerHeight, outerWidth, outerHeight, margin) {
        return (outerWidth / outerHeight > innerWidth / innerHeight ? Sizing.FixedHeight : Sizing.FixedWidth)(innerWidth, innerHeight, outerWidth, outerHeight, margin);
      }
    };

    var Layer =
    /*#__PURE__*/
    function () {
      function Layer(options) {
        this.canvas = _document.createElement('canvas');
        this.context = this.canvas.getContext('2d');

        _assign(this, Layer.defaults, options);

        if (!this.offset) {
          this.offset = new Vector();
        }

        if (this.objects) {
          this.expired = true;
        } else {
          this.expired = false;
          this.objects = [];
        }

        this.resize(this.width, this.height);
      }

      var _proto = Layer.prototype;

      _proto.resize = function resize(width, height, ratio) {
        if (ratio) {
          this.ratio = ratio;
        } else {
          ratio = this.ratio;
        }

        var canvas = this.canvas,
            offset = this.offset,
            origin = this.origin,
            originX = this.originX = width * origin.x,
            originY = this.originY = height * origin.y;
        canvas.width = (this.width = width) * ratio;
        canvas.height = (this.height = height) * ratio;
        this.context.setTransform(ratio, 0, 0, ratio, (originX - offset.x) * ratio, (originY - offset.y) * ratio);
      };

      _proto.cache = function cache(immediate) {
        var _this = this;

        if (immediate) {
          this.expired = false;
          this.fill(this.background);
          this.objects.forEach(function (object) {
            object.render(_this);
          });
        } else {
          this.expired = true;
        }
      };

      _proto.fill = function fill(color) {
        var context = this.context,
            originX = this.originX,
            originY = this.originY,
            width = this.width,
            height = this.height;

        if (color) {
          context.fillStyle = color;
          context.fillRect(-originX, -originY, width, height);
        } else {
          context.clearRect(-originX, -originY, width, height);
        }
      };

      _proto.update = function update(timeScale) {
        if (this.active) {
          this.objects.filter(function (object) {
            if (object.update) {
              if (object.defer) {
                return true;
              } else {
                object.update(timeScale);
              }
            }
          }).forEach(function (object) {
            object.update(timeScale);
          });
        }

        if (this.expired) {
          this.cache(true);
        }
      };

      _proto.render = function render(renderer) {
        Style.Common.apply(renderer, Style.Common.defaults);
        var offset = this.offset;
        renderer.context.drawImage(this.canvas, offset.x - this.originX, offset.y - this.originY, this.width, this.height);
      };

      return Layer;
    }();
    Layer.defaults = {
      active: true,
      width: 480,
      height: 320,
      ratio: Utils.Const.IS_TOUCH_MODE ? _window.devicePixelRatio || 1 : 2,
      origin: Vector.of(.5, .5),
      background: _null
    };

    var Renderer =
    /*#__PURE__*/
    function (_EventEmitter) {
      _inheritsLoose(Renderer, _EventEmitter);

      function Renderer(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _EventEmitter.call(this) || this;
        _this.settings = Utils.Const.EMPTY_OBJECT;
        _this.bounds = new Bounds();
        _this.resizeListener = Utils.debounce(_this._resizeListener.bind(_assertThisInitialized(_this)));

        _assign(_assertThisInitialized(_this), Renderer.defaults, options);

        var _assertThisInitialize = _assertThisInitialized(_this),
            canvas = _assertThisInitialize.canvas;

        if (!canvas) {
          canvas = _this.canvas || (_this.canvas = _document.createElement('canvas'));
        } else if (canvas.parentNode) {
          _this.parent = canvas.parentElement;
        }

        if (_this.parent) {
          _this.parent.appendChild(canvas);
        }

        _this.context = canvas.getContext('2d', _this.settings);

        var _assertThisInitialize2 = _assertThisInitialized(_this),
            resizeListener = _assertThisInitialize2.resizeListener;

        _this.resizeEvents.forEach(function (event) {
          _window.addEventListener(event, resizeListener);
        });

        _this._width = _this.width;
        _this._height = _this.height;

        _this._resize();

        return _this;
      }

      var _proto = Renderer.prototype;

      _proto._resize = function _resize() {
        var canvas = this.canvas,
            ratio = this.ratio,
            origin = this.origin,
            style = canvas.style;
        var _width = this._width,
            _height = this._height;
        var width = _width,
            height = _height,
            styleWidth = _width,
            styleHeight = _height;

        if (this.align) {
          var parent = this.parent;

          if (parent) {
            var rect = parent.getBoundingClientRect(),
                size = this.sizing(_width, _height, rect.width, rect.height, this.margin);
            style.marginLeft = size.left + 'px';
            style.marginTop = size.top + 'px';
            width = this.width = size.width;
            height = this.height = size.height;
            styleWidth = size.styleWidth;
            styleHeight = size.styleHeight;
            this._offsetX = rect.left + size.left;
            this._offsetY = rect.top + size.top;
            this._scale = size.scale;
          }
        } else {
          this._scale = 1;
        }

        style.width = styleWidth + 'px';
        style.height = styleHeight + 'px';
        canvas.width = width * ratio;
        canvas.height = height * ratio;
        this.emit('resize');
        var originX = this.originX = width * origin.x,
            originY = this.originY = height * origin.y;
        this.context.setTransform(ratio, 0, 0, ratio, originX * ratio, originY * ratio);
        var bounds = this.bounds;
        bounds.top = -originY;
        bounds.right = -originX + width;
        bounds.bottom = -originY + height;
        bounds.left = -originX;
      };

      _proto._resizeListener = function _resizeListener() {
        this._resize();
      };

      _proto.resize = function resize(width, height, ratio) {
        this._width = this.width = width;
        this._height = this.height = height;

        if (ratio) {
          this.ratio = ratio;
        }

        this._resize();

        return this;
      };

      _proto.fill = function fill(color) {
        var context = this.context,
            bounds = this.bounds,
            left = bounds.left,
            top = bounds.top,
            width = bounds.width,
            height = bounds.height;

        if (color) {
          context.fillStyle = color;
          context.fillRect(left, top, width, height);
        } else {
          context.clearRect(left, top, width, height);
        }
      };

      _proto.render = function render(renderable) {
        var context = this.context,
            restoration = this.restoration;

        if (restoration) {
          context.save();
        }

        renderable.render(this);

        if (restoration) {
          context.restore();
        }
      };

      _proto.toInnerPosition = function toInnerPosition(position) {
        var _scale = this._scale;
        return Vector.of((position.x - this._offsetX) / _scale - this.originX, (position.y - this._offsetY) / _scale - this.originY);
      };

      _proto.toOuterPosition = function toOuterPosition(position) {
        var _scale = this._scale;
        return Vector.of((position.x + this.originX) * _scale + this._offsetX, (position.y + this.originY) * _scale + this._offsetY);
      };

      _proto.createLayer = function createLayer(options) {
        var bounds = this.bounds;
        return new Layer(_assign({
          width: bounds.width,
          height: bounds.height,
          ratio: this.ratio
        }, options));
      };

      _createClass(Renderer, [{
        key: "resizeDelay",
        set: function set(delay) {
          this.resizeListener.delay = delay;
        },
        get: function get() {
          return this.resizeListener.delay;
        }
      }]);

      return Renderer;
    }(EventEmitter);
    Renderer.defaults = {
      settings: Utils.Const.EMPTY_OBJECT,
      width: Layer.defaults.width,
      height: Layer.defaults.height,
      margin: 0,
      ratio: Layer.defaults.ratio,
      origin: Layer.defaults.origin,
      parent: _document.body,
      align: true,
      sizing: Sizing.Fit,
      resizeEvents: ['resize', 'orientationchange'],
      resizeDelay: 100,
      restoration: false
    };

    var Sprite =
    /*#__PURE__*/
    function () {
      function Sprite(options) {
        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _assign(this, options);

        if (!this.position) {
          this.position = new Vector();
        }

        this.style = _assign({}, Sprite.defaultStyle, this.style);
      }

      Sprite.of = function of(src, options) {
        return new Sprite(options).load(src);
      };

      var _proto = Sprite.prototype;

      _proto.load = function load(src, onSuccess, onFailure) {
        var image = this.image = _document.createElement('img');

        image.src = src;

        if (onSuccess) {
          image.onload = onSuccess;
        }

        if (onFailure) {
          image.onabort = image.onerror = onFailure;
        }

        return this;
      };

      _proto.render = function render(renderer) {
        if (!this.visible) {
          return;
        }

        var image = this.image;

        if (image) {
          var position = this.position,
              rotation = this.rotation,
              width = image.width,
              height = image.height,
              context = renderer.context,
              dstW = this.width || width,
              dstH = this.height || height;
          Style.Common.apply(renderer, this.style);
          context.rotate(rotation);
          context.drawImage(image, (this.clipX || 0) - dstW / 2, (this.clipY || 0) - dstH / 2, this.clipWidth || width, this.clipHeight || height, position.x, position.y, dstW, dstH);
          context.rotate(-rotation);
        }
      };

      return Sprite;
    }();
    Sprite.defaults = {
      visible: true,
      rotation: 0
    };
    Sprite.defaultStyle = _assign({}, Style.Common.defaults);

    var Shape =
    /*#__PURE__*/
    function (_Body) {
      _inheritsLoose(Shape, _Body);

      function Shape(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _Body.call(this, _assign({}, Shape.defaults, options)) || this;
        _this.style = _assign({}, Shape.defaultStyle, _this.style);

        if (!_this.attachments) {
          _this.attachments = [];
        }

        return _this;
      }

      Shape.applyStyle = function applyStyle(renderer, style) {
        Style.Common.apply(renderer, style);
        Style.Fill.apply(renderer, style);
        Style.Stroke.apply(renderer, style);
      };

      var _proto = Shape.prototype;

      _proto.render = function render(renderer) {
        if (!this.visible) {
          return;
        }

        var style = this.style,
            fillFirst = this.fillFirst,
            position = this.position,
            texture = this.texture,
            attachments = this.attachments,
            fillStyle = style.fillStyle,
            context = renderer.context,
            TRANSPARENT = Utils.Const.TRANSPARENT;
        context.translate(position.x, position.y);

        if (texture) {
          var rotation = this.rotation;
          context.rotate(rotation);
          texture.render(renderer);
          context.rotate(-rotation);
        } else {
          if (this.preferShadow && !style.shadowBlur && style.shadowColor !== TRANSPARENT) {
            context.shadowColor = TRANSPARENT;
            var shadowOffsetX = style.shadowOffsetX,
                shadowOffsetY = style.shadowOffsetY;
            context.translate(shadowOffsetX, shadowOffsetY);
            context.beginPath();
            this.path(context);

            if (this.closePath) {
              context.closePath();
            }

            context.fillStyle = style.shadowColor;
            context.fill();
            context.translate(-shadowOffsetX, -shadowOffsetY);

            if (fillStyle) {
              context.fillStyle = fillStyle;
            }
          }

          Shape.applyStyle(renderer, style);
          context.beginPath();
          this.path(context);

          if (this.closePath) {
            context.closePath();
          }

          if (fillFirst && fillStyle) {
            context.fill();
            context.shadowColor = TRANSPARENT;
          }

          if (style.strokeStyle) {
            context.stroke();
            context.shadowColor = TRANSPARENT;
          }

          if (!fillFirst && fillStyle) {
            context.fill();
          }

          if (attachments.length) {
            attachments.forEach(function (attachment) {
              attachment.render(renderer);
            });
          }
        }

        context.translate(-position.x, -position.y);
      };

      return Shape;
    }(Body);
    Shape.defaults = {
      visible: true,
      fillFirst: true,
      closePath: true,
      preferShadow: false,
      texture: _null
    };
    Shape.defaultStyle = _assign({}, Style.Common.defaults, Style.Stroke.defaults, {
      fillStyle: _null
    });

    var Text =
    /*#__PURE__*/
    function () {
      function Text(options) {
        _assign(this, Text.defaults, options);

        if (!this.position) {
          this.position = new Vector();
        }

        this.style = _assign({}, Text.defaultStyle, this.style);
      }

      Text.applyStyle = function applyStyle(renderer, textStyle) {
        Shape.applyStyle(renderer, textStyle);
        var context = renderer.context;
        context.font = textStyle.font;
        context.textAlign = textStyle.textAlign;
        context.textBaseline = textStyle.textBaseline;
        context.direction = textStyle.direction;
      };

      var _proto = Text.prototype;

      _proto.render = function render(renderer) {
        var style = this.style,
            fillFirst = this.fillFirst,
            content = this.content,
            position = this.position,
            rotation = this.rotation,
            x = position.x,
            y = position.y,
            fillStyle = style.fillStyle,
            context = renderer.context,
            TRANSPARENT = Utils.Const.TRANSPARENT;
        Text.applyStyle(renderer, style);
        context.translate(x, y);
        context.rotate(rotation);

        if (this.preferShadow && !style.shadowBlur && style.shadowColor !== TRANSPARENT) {
          context.shadowColor = TRANSPARENT;
          context.fillStyle = style.shadowColor;
          context.fillText(content, style.shadowOffsetX, style.shadowOffsetY);

          if (fillStyle) {
            context.fillStyle = fillStyle;
          }
        }

        if (fillFirst && fillStyle) {
          context.fillText(content, 0, 0);
          context.shadowColor = TRANSPARENT;
        }

        if (style.strokeStyle) {
          context.strokeText(content, 0, 0);
          context.shadowColor = TRANSPARENT;
        }

        if (!fillFirst && fillStyle) {
          context.fillText(content, 0, 0);
        }

        context.rotate(-rotation);
        context.translate(-x, -y);
      };

      return Text;
    }();
    Text.defaults = {
      visible: true,
      content: '',
      preferShadow: false,
      rotation: 0
    };
    Text.defaultStyle = _assign({}, Shape.defaultStyle, {
      font: '16px Consolas',
      textAlign: 'center',
      textBaseline: 'middle',
      direction: 'inherit'
    });

    var Paragraph =
    /*#__PURE__*/
    function () {
      function Paragraph(options) {
        _assign(this, Paragraph.defaults, options);

        if (!this.position) {
          this.position = new Vector();
        }

        this.style = _assign({}, Paragraph.defaultStyle, this.style);
      }

      var _proto = Paragraph.prototype;

      _proto.render = function render(renderer) {
        var style = this.style,
            fillFirst = this.fillFirst,
            lines = this.lines,
            lineHeight = this.lineHeight,
            indent = this.indent,
            position = this.position,
            rotation = this.rotation,
            fillStyle = style.fillStyle,
            strokeStyle = style.strokeStyle,
            shadowColor = style.shadowColor,
            shadowOffsetX = style.shadowOffsetX,
            shadowOffsetY = style.shadowOffsetY,
            x = position.x,
            y = position.y,
            context = renderer.context,
            TRANSPARENT = Utils.Const.TRANSPARENT,
            preferShadow = this.preferShadow && !style.shadowBlur && style.shadowColor !== TRANSPARENT;
        Text.applyStyle(renderer, style);

        if (preferShadow) {
          context.shadowColor = TRANSPARENT;
        }

        context.translate(x, y);
        context.rotate(rotation);
        var dx = 0,
            dy = 0;
        lines.forEach(function (line) {
          if (preferShadow) {
            context.fillStyle = style.shadowColor;
            context.fillText(line, dx + shadowOffsetX, dy + shadowOffsetY);

            if (fillStyle) {
              context.fillStyle = fillStyle;
            }
          }

          if (fillFirst && fillStyle) {
            context.fillText(line, dx, dy);
            context.shadowColor = TRANSPARENT;
          }

          if (strokeStyle) {
            context.strokeText(line, dx, dy);
            context.shadowColor = TRANSPARENT;
          }

          if (!fillFirst && fillStyle) {
            context.fillText(line, dx, dy);
          }

          context.shadowColor = shadowColor;
          dx += indent;
          dy += lineHeight;
        });
        context.rotate(-rotation);
        context.translate(-x, -y);
      };

      return Paragraph;
    }();
    Paragraph.defaults = {
      visible: true,
      lines: [],
      indent: 0,
      lineHeight: 20,
      preferShadow: true,
      rotation: 0
    };
    Paragraph.defaultStyle = _assign({}, Text.defaultStyle);

    var Rectangle =
    /*#__PURE__*/
    function (_Shape) {
      _inheritsLoose(Rectangle, _Shape);

      function Rectangle(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _Shape.call(this, _assign({}, Rectangle.defaults, options)) || this;

        _this.updateBounds();

        var _assertThisInitialize = _assertThisInitialized(_this),
            rotation = _assertThisInitialize.rotation,
            cos = _cos(rotation),
            sin = _sin(rotation);

        _this.normals = [Vector.of(cos, sin), Vector.of(-sin, cos)];

        _this.updateSize();

        return _this;
      }

      var _proto = Rectangle.prototype;

      _proto._scale = function _scale(scaleX, scaleY, origin) {
        this.width *= scaleX;
        this.height *= scaleY;
        var _this$position = this.position,
            x0 = _this$position.x,
            y0 = _this$position.y,
            bounds = this.bounds,
            x = origin ? origin.x + (x0 - origin.x) * scaleX : x0,
            y = origin ? origin.y + (y0 - origin.y) * scaleY : y0,
            dx = (bounds.right - x) * scaleX,
            dy = (bounds.bottom - y) * scaleY;
        bounds.left = x - dx;
        bounds.right = x + dx;
        bounds.top = y - dy;
        bounds.bottom = y + dy;
      };

      _proto._rotate = function _rotate(rotation, origin) {
        var normals = this.normals;
        normals[0].rotate(rotation);
        normals[1].rotate(rotation);
        this.updateBounds();
      };

      _proto.updateSize = function updateSize(width, height) {
        if (width === _undefined) {
          width = this.width;
        } else {
          this.width = width;
        }

        if (height === _undefined) {
          height = this.height;
        } else {
          this.height = height;
        }

        this._setArea(width * height);

        this.updateBounds();
      };

      _proto.updateBounds = function updateBounds() {
        var bounds = this.bounds,
            rotation = this.rotation,
            radius = this.radius,
            width = this.width,
            height = this.height,
            position = this.position,
            x = position.x,
            y = position.y;
        var dx0 = width / 2,
            dy0 = height / 2;

        if (radius > 0) {
          dx0 -= radius;
          dy0 -= radius;
        }

        var cos = _cos(rotation),
            sin = _sin(rotation),
            dx1 = dx0 * cos - dy0 * sin,
            dy1 = dx0 * sin + dy0 * cos,
            dx2 = dx0 * cos - -dy0 * sin,
            dy2 = dx0 * sin + -dy0 * cos;

        bounds.updateVertices(Vertices.fromArray([x + dx1, y + dy1, x + dx2, y + dy2, x - dx1, y - dy1, x - dx2, y - dy2]));

        if (radius > 0) {
          bounds.grow(radius);
        }
      };

      _proto.getClosest = function getClosest(target) {
        var rotation = this.rotation,
            radius = this.radius,
            width = this.width,
            height = this.height,
            position = this.position,
            x = position.x,
            y = position.y;
        var dx0 = width / 2,
            dy0 = height / 2;

        if (radius > 0) {
          dx0 -= radius;
          dy0 -= radius;
        }

        var cos = _cos(rotation),
            sin = _sin(rotation),
            dx1 = dx0 * cos - dy0 * sin,
            dy1 = dx0 * sin + dy0 * cos,
            dx2 = dx0 * cos - -dy0 * sin,
            dy2 = dx0 * sin + -dy0 * cos,
            closest = Vertices.findClosest(target, Vertices.fromArray([x + dx1, y + dy1, x + dx2, y + dy2, x - dx1, y - dy1, x - dx2, y - dy2]));

        if (radius > 0) {
          closest.grow(radius);
        }

        return closest;
      };

      _proto.project = function project(direction) {
        var radius = this.radius,
            angle = this.rotation - direction.getAngle(),
            cos = _cos(angle),
            sin = _sin(angle);

        var x0 = this.width / 2,
            y0 = this.height / 2;

        if (radius > 0) {
          x0 -= radius;
          y0 -= radius;
        }

        var halfLength = _max(_abs(x0 * cos - y0 * sin), _abs(x0 * cos + y0 * sin));

        if (radius > 0) {
          halfLength += radius;
        }

        var positionProjection = Vector.project(this.position, direction);
        return {
          min: positionProjection - halfLength,
          max: positionProjection + halfLength
        };
      };

      _proto.path = function path(context) {
        var width = this.width,
            height = this.height,
            radius = this.radius,
            rotation = this.rotation;
        context.rotate(rotation);

        if (radius > 0) {
          var halfWidth = width / 2,
              halfHeight = height / 2,
              innerHalfWidth = halfWidth - radius,
              innerHalfHeight = halfHeight - radius;
          context.moveTo(-innerHalfWidth, -halfHeight);
          context.arcTo(halfWidth, -halfHeight, halfWidth, innerHalfHeight, radius);
          context.arcTo(halfWidth, halfHeight, -innerHalfWidth, halfHeight, radius);
          context.arcTo(-halfWidth, halfHeight, -halfWidth, -innerHalfHeight, radius);
          context.arcTo(-halfWidth, -halfHeight, -innerHalfWidth, -halfHeight, radius);
        } else {
          context.rect(-width / 2, -height / 2, width, height);
        }

        context.rotate(-rotation);
      };

      return Rectangle;
    }(Shape);
    Rectangle.defaults = {
      width: 20,
      height: 20,
      radius: 0
    };

    var Circle =
    /*#__PURE__*/
    function (_Shape) {
      _inheritsLoose(Circle, _Shape);

      function Circle(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _Shape.call(this, _assign({}, Circle.defaults, options)) || this;
        _this.isCircle = true;

        if (_this.radius) {
          _this.updateRadius();
        }

        return _this;
      }

      var _proto = Circle.prototype;

      _proto.updateRadius = function updateRadius(radius) {
        if (radius === _undefined) {
          radius = this.radius;
        } else {
          this.radius = radius;
        }

        this._setArea(radius * radius * _PI);

        this.updateBounds();
      };

      _proto.updateBounds = function updateBounds() {
        var bounds = this.bounds,
            _this$position = this.position,
            x = _this$position.x,
            y = _this$position.y,
            radius = this.radius,
            rotation = this.rotation,
            cos = _cos(rotation),
            sin = -_sin(rotation),
            a = radius * this.scaleX,
            b = radius * this.scaleY,
            halfWidth = _sqrt(Utils.quadraticSum(a * cos, b * sin)),
            halfHeight = _sqrt(Utils.quadraticSum(b * cos, a * sin));

        bounds.left = x - halfWidth;
        bounds.right = x + halfWidth;
        bounds.top = y - halfHeight;
        bounds.bottom = y + halfHeight;
      };

      _proto.getClosest = function getClosest(target) {
        var position = this.position,
            deltaVector = Vector.minus(target, position),
            angle = deltaVector.getAngle();
        return deltaVector.setNorm(this.radius * _sqrt(Utils.quadraticSum(_cos(angle) * this.scaleX, _sin(angle) * this.scaleY))).plusVector(position);
      };

      _proto.project = function project(direction) {
        var positionProjection = Vector.project(this.position, direction),
            angle = this.rotation - direction.getAngle(),
            halfLength = this.radius * _sqrt(Utils.quadraticSum(_cos(angle) * this.scaleX, _sin(angle) * this.scaleY));

        return {
          min: positionProjection - halfLength,
          max: positionProjection + halfLength
        };
      };

      _proto.path = function path(context) {
        var rotation = this.rotation,
            scaleX = this.scaleX,
            scaleY = this.scaleY;
        context.rotate(rotation);
        context.scale(scaleX, scaleY);
        context.arc(0, 0, this.radius, 0, Utils.Const.DOUBLE_PI);
        context.rotate(-rotation);
        context.scale(1 / scaleX, 1 / scaleY);
      };

      return Circle;
    }(Shape);
    Circle.defaults = {
      radius: 1
    };

    var Polygon =
    /*#__PURE__*/
    function (_Shape) {
      _inheritsLoose(Polygon, _Shape);

      function Polygon(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _Shape.call(this, _assign({}, Polygon.defaults, options)) || this;

        if (_this.vertices) {
          _this.updateVertices();
        } else {
          _this.vertices = [];
        }

        return _this;
      }

      var _proto = Polygon.prototype;

      _proto.updateVertices = function updateVertices(vertices) {
        if (vertices) {
          this.vertices = vertices;
        } else {
          vertices = this.vertices;
        }

        if (vertices.length > 1) {
          var clockwise = this.clockwise,
              adjustment = this.adjustment,
              normalPrecision = Body.normalPrecision,
              areas = new Array(),
              normals = new Array(),
              centers = new Array(),
              tangents = new _Set();
          var totalArea = 0;
          vertices.reduce(function (vertex1, vertex2) {
            var area = Vector.cross(vertex1, vertex2) / 2;
            totalArea += area;
            var normal = Vector.minus(vertex2, vertex1).turn(clockwise),
                tangent = (normal.y / normal.x).toFixed(normalPrecision);

            if (!tangents.has(tangent)) {
              tangents.add(tangent);
              normals.push(normal.normalize());
            }

            if (adjustment) {
              centers.push(Vector.of((vertex1.x + vertex2.x) / 3, (vertex1.y + vertex2.y) / 3));
              areas.push(area);
            }

            return vertex2;
          }, vertices[vertices.length - 1]);

          if (adjustment) {
            var offset = Vector.plus(centers.map(function (center, i) {
              return center.scale(areas[i] / totalArea);
            }));
            vertices.forEach(function (vertex) {
              vertex.minusVector(offset);
            });
          }

          if (!clockwise) {
            totalArea = -totalArea;
          }

          this.normals = normals;

          this._setArea(totalArea);
        } else {
          this._setArea(0);
        }

        this.updateBounds();
        return this;
      };

      _proto._scale = function _scale(scaleX, scaleY, origin) {
        this.vertices.forEach(function (vertex) {
          vertex.scale(scaleX, scaleY, origin);
        });
      };

      _proto._rotate = function _rotate(rotation, origin) {
        this.vertices.forEach(function (vertex) {
          vertex.rotate(rotation, origin);
        });
        this.normals.forEach(function (normal) {
          normal.rotate(rotation, origin);
        });
      };

      _proto.updateBounds = function updateBounds() {
        var vertices = this.vertices,
            bounds = this.bounds,
            position = this.position;

        if (vertices.length) {
          bounds.updateVertices(vertices);
          bounds.moveVector(position);
        } else {
          bounds.left = bounds.right = position.x;
          bounds.top = bounds.bottom = position.y;
        }
      };

      _proto.getClosest = function getClosest(target) {
        return Vertices.findClosest(target, this.vertices);
      };

      _proto.project = function project(direction) {
        var vertices = this.vertices;
        var min = 0,
            max = 0,
            projection;

        if (vertices.length) {
          vertices.forEach(function (vertex, i) {
            projection = Vector.project(vertex, direction);

            if (i > 0) {
              if (projection < min) {
                min = projection;
              } else if (projection > max) {
                max = projection;
              }
            } else {
              min = max = projection;
            }
          });
        }

        var positionProjection = Vector.project(this.position, direction);
        return {
          min: min + positionProjection,
          max: max + positionProjection
        };
      };

      _proto.path = function path(context) {
        this.vertices.forEach(function (vertex, i) {
          if (i > 0) {
            context.lineTo(vertex.x, vertex.y);
          } else {
            context.moveTo(vertex.x, vertex.y);
          }
        });
      };

      return Polygon;
    }(Shape);
    Polygon.defaults = {
      clockwise: true,
      adjustment: true
    };

    var Line =
    /*#__PURE__*/
    function (_Shape) {
      _inheritsLoose(Line, _Shape);

      function Line(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _Shape.call(this, _assign({}, Line.defaults, options)) || this;

        if (!_this.start) {
          _this.start = new Vector();
        }

        if (!_this.end) {
          _this.end = new Vector();
        }

        _this.updateVertices();

        return _this;
      }

      var _proto = Line.prototype;

      _proto._scale = function _scale(scaleX, scaleY, origin) {
        this.start.scale(scaleX, scaleY, origin);
        this.end.scale(scaleX, scaleY, origin);
        this.updateVertices();
      };

      _proto._rotate = function _rotate(rotation, origin) {
        this.start.rotate(rotation, origin);
        this.end.rotate(rotation, origin);
        this.updateVertices();
      };

      _proto.updateVertices = function updateVertices(start, end) {
        if (start) {
          this.start = start;
        }

        if (end) {
          this.end = end;
        }

        var deltaVector = Vector.minus(this.end, this.start);
        this.normals = [deltaVector, Vector.of(-deltaVector.y, deltaVector.x)];

        this._setArea(deltaVector.getNorm());

        this.updateBounds();
      };

      _proto.updateBounds = function updateBounds() {
        var _this$start = this.start,
            x1 = _this$start.x,
            y1 = _this$start.y,
            _this$end = this.end,
            x2 = _this$end.x,
            y2 = _this$end.y,
            bounds = this.bounds;

        if (x1 > x2) {
          bounds.left = x2;
          bounds.right = x1;
        } else {
          bounds.left = x1;
          bounds.right = x2;
        }

        if (y1 > y2) {
          bounds.top = y2;
          bounds.bottom = y1;
        } else {
          bounds.top = y1;
          bounds.bottom = y2;
        }
      };

      _proto.getClosest = function getClosest(target) {
        var start = this.start,
            end = this.end;
        return Utils.quadraticSum(start.x - target.x, start.y - target.y) < Utils.quadraticSum(end.x - target.x, end.y - target.y) ? start : end;
      };

      _proto.project = function project(direction) {
        var start = this.start,
            end = this.end,
            startProjection = Vector.project(start, direction),
            endProjection = Vector.project(end, direction);
        return startProjection < endProjection ? {
          min: startProjection,
          max: endProjection
        } : {
          min: endProjection,
          max: startProjection
        };
      };

      _proto.path = function path(context) {
        var start = this.start,
            end = this.end;
        context.moveTo(start.x, start.y);
        context.lineTo(end.x, end.y);
      };

      return Line;
    }(Shape);
    Line.defaults = {
      closePath: false
    };

    var Particles =
    /*#__PURE__*/
    function () {
      function Particles(options) {
        this._items = [];
        this._deadlines = new _Map();
        this._legacies = new _Map();
        this.spawn = Utils.throttle(this._spawn);

        _assign(this, Particles.defaults, options);

        if (!this.offset) {
          this.offset = new Vector();
        }

        if (!this.items) {
          this.items = [];
        }
      }

      var _proto = Particles.prototype;

      _proto._spawn = function _spawn() {
        var pool = this.pool;

        if (!pool) {
          return;
        }

        var items = this.items,
            _deadlines = this._deadlines,
            initializer = this.initializer,
            particle = pool.get(),
            life = Utils.mix(this.minLife, this.maxLife, _Math.random());

        _deadlines.set(particle, _now() + life);

        if (initializer) {
          this._legacies.set(particle, initializer(particle, life));
        }

        this._items.push(particle);

        if (this.sequential) {
          items.push(particle);
        } else {
          Utils.insert(items, _Math.round(items.length * _Math.random()), particle);
        }
      };

      _proto.clear = function clear() {
        var pool = this.pool,
            cleaner = this.cleaner,
            _legacies = this._legacies;
        this.items.forEach(function (item) {
          if (pool) {
            pool.add(item);
          }

          if (cleaner) {
            cleaner(item, _legacies.get(item));
          }
        });
        this.items = [];
        this._items.length = 0;

        this._deadlines.clear();

        this._legacies.clear();
      };

      _proto._remove = function _remove(item) {
        var _deadlines = this._deadlines,
            pool = this.pool,
            cleaner = this.cleaner,
            _legacies = this._legacies,
            _items = this._items;
        Utils.removeIndex(_items, _items.indexOf(item));

        _deadlines["delete"](item);

        if (pool) {
          pool.add(item);
        }

        if (cleaner) {
          cleaner(item, _legacies.get(item));
        }

        _legacies["delete"](item);
      };

      _proto.update = function update(timeScale) {
        var _this = this;

        if (!this.active) {
          return;
        }

        var _deadlines = this._deadlines,
            now = _now();

        this.items = this.items.filter(function (item) {
          if (_deadlines.get(item) <= now) {
            return _this._remove(item);
          }

          if (item.update) {
            item.update(timeScale);
          }

          return true;
        });

        if (this.loop) {
          this.spawn();
        }

        var currentCount = this.items.length;

        if (currentCount > this.maxCount) {
          var items = this.items;

          if (this.sequential) {
            for (var i = this.maxCount; i < currentCount; i++) {
              this._remove(items.shift());
            }
          } else {
            var _items = this._items;

            for (var _i = this.maxCount; _i < currentCount; _i++) {
              var item = _items[0];

              this._remove(item);

              Utils.removeIndex(items, items.indexOf(item));
            }
          }
        } else if (currentCount < this.minCount) {
          for (var _i2 = currentCount; _i2 < this.minCount; _i2++) {
            this._spawn();
          }
        }
      };

      _proto.render = function render(renderer) {
        if (!this.visible) {
          return;
        }

        var items = this.items,
            offset = this.offset,
            context = renderer.context;
        context.translate(offset.x, offset.y);
        items.forEach(function (item) {
          item.render(renderer);
        });
        context.translate(-offset.x, -offset.y);
      };

      _createClass(Particles, [{
        key: "life",
        set: function set(value) {
          this.minLife = this.maxLife = value;
        }
      }, {
        key: "count",
        set: function set(value) {
          this.minCount = this.maxCount = value;
        }
      }, {
        key: "threshold",
        set: function set(value) {
          this.spawn.threshold = value;
        },
        get: function get() {
          return this.spawn.threshold;
        }
      }]);

      return Particles;
    }();
    Particles.defaults = {
      visible: true,
      active: true,
      pool: _null,
      minCount: 0,
      maxCount: _Infinity,
      minLife: _Infinity,
      maxLife: _Infinity,
      initializer: _null,
      cleaner: _null,
      sequential: false,
      loop: true
    };

    var Timer =
    /*#__PURE__*/
    function (_EventEmitter) {
      _inheritsLoose(Timer, _EventEmitter);

      function Timer(options) {
        var _this;

        _this = _EventEmitter.call(this) || this;
        _this.lastFrameDelay = 0;
        _this.lastFrameDuration = 0;
        _this.isRunning = false;
        _this._timeoutId = 0;
        _this._timeoutCallbacks = new Map();
        _this._intervalId = 0;
        _this._intervalCallbacks = new Map();

        _assign(_assertThisInitialized(_this), Timer.defaults, options);

        _this._tick = _this._tick.bind(_assertThisInitialized(_this));
        return _this;
      }

      var _proto = Timer.prototype;

      _proto._requestTick = function _requestTick(delay) {
        this._timer = (this._usedRAF = this.delay <= Timer.RAFThreshold && this.allowRAF) ? _window.requestAnimationFrame(this._tick) : _setTimeout(this._tick, _max(delay, 0));
      };

      _proto._tick = function _tick() {
        var startTime = _now(),
            deltaTime = this.lastFrameDelay = startTime - this._lastTickTime,
            _timer = this._timer;

        this._lastTickTime = startTime;
        var _timeoutCallbacks = this._timeoutCallbacks;

        _timeoutCallbacks.forEach(function (record, id) {
          if (startTime >= record[1]) {
            record[0].apply(_undefined, record[2]);

            _timeoutCallbacks["delete"](id);
          }
        });

        this._intervalCallbacks.forEach(function (record) {
          if (startTime >= record[1]) {
            record[0].apply(_undefined, record[3]);

            while (startTime >= record[1]) {
              record[1] += record[2];
            }
          }
        });

        this.emit('tick', deltaTime);

        if (!this.isRunning) {
          return;
        }

        var duration = this.lastFrameDuration = _now() - startTime;

        if (this._timer === _timer) {
          this._requestTick(_max(this.delay - (this.fixDelay ? duration : 0), 0));
        }
      };

      _proto.start = function start() {
        this.isRunning = true;

        var now = this._lastTickTime = _now(),
            _lastStopTime = this._lastStopTime;

        if (_lastStopTime !== _undefined) {
          var delay = now - _lastStopTime;

          this._timeoutCallbacks.forEach(function (record) {
            record[1] += delay;
          });

          this._intervalCallbacks.forEach(function (record) {
            record[1] += delay;
          });
        }

        this._tick();
      };

      _proto.stop = function stop() {
        if (!this.isRunning) {
          return;
        }

        this.isRunning = false;
        this._lastStopTime = _now();
        var _timer = this._timer;

        if (_timer !== _undefined) {
          (this._usedRAF ? _window.cancelAnimationFrame : _clearTimeout)(_timer);
          this._timer = _undefined;
        }
      };

      _proto.reschedule = function reschedule(delay) {
        this.delay = delay;

        if (delay < 0) {
          return this.stop();
        }

        if (this.isRunning) {
          this.stop();
          this.isRunning = true;

          var now = _now();

          this._requestTick(_max(delay - (now - this._lastTickTime), 0));

          this._lastTickTime = now;
        }
      };

      _proto.setTimeout = function setTimeout(callback, timeout) {
        var id = this._timeoutId++;

        for (var _len = arguments.length, args = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
          args[_key - 2] = arguments[_key];
        }

        this._timeoutCallbacks.set(id, [callback, _now() + timeout, args]);

        return id;
      };

      _proto.clearTimeout = function clearTimeout(id) {
        this._timeoutCallbacks["delete"](id);
      };

      _proto.setInterval = function setInterval(callback, interval) {
        var id = this._intervalId++;

        for (var _len2 = arguments.length, args = new Array(_len2 > 2 ? _len2 - 2 : 0), _key2 = 2; _key2 < _len2; _key2++) {
          args[_key2 - 2] = arguments[_key2];
        }

        this._intervalCallbacks.set(id, [callback, _now() + interval, interval, args]);

        return id;
      };

      _proto.clearInterval = function clearInterval(id) {
        this._intervalCallbacks["delete"](id);
      };

      return Timer;
    }(EventEmitter);
    Timer.RAFThreshold = 1000 / 60;
    Timer.defaults = {
      delay: Timer.RAFThreshold,
      allowRAF: true,
      fixDelay: true
    };

    var Pointer =
    /*#__PURE__*/
    function (_EventEmitter) {
      _inheritsLoose(Pointer, _EventEmitter);

      function Pointer(options) {
        var _this;

        _this = _EventEmitter.call(this) || this;
        _this.isCircle = true;
        _this.position = new Vector();
        _this.bounds = new Bounds();
        _this.normals = [];
        _this.active = true;
        _this.isHolding = false;
        _this.startTimeStamps = new _Map();
        _this.positions = new _Map();

        _assign(_assertThisInitialized(_this), Pointer.defaults, options);

        var _assertThisInitialize = _assertThisInitialized(_this),
            target = _assertThisInitialize.target;

        if (_this.isTouchMode) {
          target.addEventListener('touchstart', _this._startListener = function (event) {
            var touchPoint = event.changedTouches[0];

            _this._start(touchPoint.identifier, Vector.of(touchPoint.clientX, touchPoint.clientY), event);
          });
          target.addEventListener('touchmove', _this._moveListener = function (event) {
            var touchPoint = event.changedTouches[0];

            _this._move(touchPoint.identifier, Vector.of(touchPoint.clientX, touchPoint.clientY), event);
          });
          target.addEventListener('touchend', _this._endListener = function (event) {
            var touchPoint = event.changedTouches[0];

            _this._end(touchPoint.identifier, Vector.of(touchPoint.clientX, touchPoint.clientY), event);
          });
        } else {
          target.addEventListener('mousedown', _this._startListener = function (event) {
            _this._start(-1, Vector.of(event.clientX, event.clientY), event);
          });
          target.addEventListener('mousemove', _this._moveListener = function (event) {
            _this._move(-1, Vector.of(event.clientX, event.clientY), event);
          });
          target.addEventListener('mouseup', _this._endListener = function (event) {
            _this._end(-1, Vector.of(event.clientX, event.clientY), event);
          });
        }

        return _this;
      }

      var _proto = Pointer.prototype;

      _proto._setPosition = function _setPosition(id, rawPosition, endFlag) {
        var bounds = this.bounds,
            radius = this.radius,
            positions = this.positions,
            pretransform = this.pretransform,
            transform = this.transform,
            tempPosition = pretransform ? pretransform(rawPosition) : rawPosition,
            newPosition = transform ? transform(tempPosition) : tempPosition;
        this.position.setVector(newPosition);

        if (endFlag) {
          positions["delete"](id);
        } else {
          positions.set(id, newPosition);
        }

        bounds.left = newPosition.x - radius;
        bounds.right = newPosition.x + radius;
        bounds.top = newPosition.y - radius;
        bounds.bottom = newPosition.y + radius;
      };

      _proto._start = function _start(id, rawPosition, event) {
        this.isHolding = true;
        this.startTimeStamps.set(id, event.timeStamp);

        this._setPosition(id, rawPosition);

        this.emit('start', this.position, id, event);
      };

      _proto._move = function _move(id, rawPosition, event) {
        if (this.holdOnly && !this.isHolding) {
          return;
        }

        this._setPosition(id, rawPosition);

        this.emit('move', this.position, id, event);
      };

      _proto._end = function _end(id, rawPosition, event) {
        this._setPosition(id, rawPosition, true);

        this.isHolding = false;
        var startTimeStamps = this.startTimeStamps,
            startTimeStamp = startTimeStamps.get(id);
        startTimeStamps["delete"](id);
        this.emit('end', this.position, id, event);

        if (event.timeStamp - startTimeStamp <= this.clickThreshold) {
          this.emit('click', this.position, id, event);
        }
      };

      _proto.getClosest = function getClosest(target) {
        return Vector.minus(target, this.position).setNorm(this.radius);
      };

      _proto.project = function project(direction) {
        var radius = this.radius,
            positionProjection = Vector.project(this.position, direction);
        return {
          min: positionProjection - radius,
          max: positionProjection + radius
        };
      };

      _proto.destroy = function destroy() {
        if (!this.active) {
          return;
        }

        this.active = false;
        var target = this.target;

        if (this.isTouchMode) {
          target.removeEventListener('touchstart', this._startListener);
          target.removeEventListener('touchmove', this._moveListener);
          target.removeEventListener('touchend', this._endListener);
        } else {
          target.removeEventListener('mousedown', this._startListener);
          target.removeEventListener('mousemove', this._moveListener);
          target.removeEventListener('mouseup', this._endListener);
        }
      };

      return Pointer;
    }(EventEmitter);
    Pointer.defaults = {
      target: _window,
      pretransform: _null,
      transform: _null,
      isTouchMode: Utils.Const.IS_TOUCH_MODE,
      clickThreshold: 1000
    };
    Pointer.defaults.radius = Pointer.defaults.isTouchMode ? 5 : 0;

    var Camera =
    /*#__PURE__*/
    function () {
      function Camera(options) {
        _assign(this, Camera.defaults, options);

        if (!this.position) {
          this.position = new Vector();
        }

        if (!this.scale) {
          this.scale = new Vector(1, 1);
        }
      }

      var _proto = Camera.prototype;

      _proto.applyTo = function applyTo(context) {
        var position = this.position,
            scale = this.scale,
            scaleX = scale.x,
            scaleY = scale.y;
        context.transform(scaleX, 0, 0, scaleY, -position.x * scaleX, -position.y * scaleY);
        context.rotate(this.rotation);
      };

      _proto.restore = function restore(context) {
        var position = this.position,
            scale = this.scale;
        context.rotate(-this.rotation);
        context.transform(1 / scale.x, 0, 0, 1 / scale.y, position.x, position.y);
      };

      _proto.toViewPosition = function toViewPosition(position) {
        var pos = this.position,
            scale = this.scale,
            rotation = this.rotation,
            cos = _cos(rotation),
            sin = -_sin(rotation),
            tx = (position.x + pos.x) / scale.x,
            ty = (position.y + pos.y) / scale.y;

        return Vector.of(tx * cos - ty * sin, tx * sin + ty * cos);
      };

      _proto.toGlobalPosition = function toGlobalPosition(position) {
        var pos = this.position,
            scale = this.scale,
            rotation = this.rotation,
            cos = _cos(rotation),
            sin = _sin(rotation),
            tx = position.x * scale.x - pos.x,
            ty = position.y * scale.y - pos.y;

        return Vector.of(tx * cos - ty * sin, tx * sin + ty * cos);
      };

      return Camera;
    }();
    Camera.defaults = {
      rotation: 0
    };

    var Scene =
    /*#__PURE__*/
    function (_EventEmitter) {
      _inheritsLoose(Scene, _EventEmitter);

      function Scene(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _EventEmitter.call(this) || this;
        _this.active = false;
        _this._pointerPosition = new Vector();
        _this.toViewPosition = _this.toViewPosition.bind(_assertThisInitialized(_this));
        _this.toGlobalPosition = _this.toGlobalPosition.bind(_assertThisInitialized(_this));
        _this._onPointerClick = _this._onPointerClick.bind(_assertThisInitialized(_this));
        _this._onPointerStart = _this._onPointerStart.bind(_assertThisInitialized(_this));
        _this._onPointerEnd = _this._onPointerEnd.bind(_assertThisInitialized(_this));

        if (!_this.pointerConstraint) {
          _this.pointerConstraint = new Constraint();
        }

        _assign(_assertThisInitialized(_this), Scene.defaults, options);

        if (!_this.objects) {
          _this.objects = [];
        }

        if (!_this.attachments) {
          _this.attachments = [];
        }

        if (!_this.effects) {
          _this.effects = [];
        }

        return _this;
      }

      var _proto = Scene.prototype;

      _proto._onPointerClick = function _onPointerClick(position, id, event) {
        if (!this.active) {
          return;
        }

        var pointerChecker = this.pointerChecker;

        if (!pointerChecker) {
          return;
        }

        var focus = this.getFocus();

        if (focus) {
          focus.emit('click', position, id, event);
        }
      };

      _proto._draggableFilter = function _draggableFilter(body) {
        return body.draggable;
      };

      _proto._onPointerStart = function _onPointerStart(position, id, event) {
        var pointerConstraint = this.pointerConstraint;

        if (!pointerConstraint) {
          return;
        }

        this._pointerPosition.setVector(position);

        if (pointerConstraint.target) {
          pointerConstraint.target.emit('dragEnd', position, id, event);
        }

        var focus = this.getFocus(this._draggableFilter);

        if (focus) {
          pointerConstraint.target = focus;
          pointerConstraint.targetOffset.setVector(position).minusVector(focus.position);
          focus.emit('dragStart', position, id, event);
        } else {
          pointerConstraint.target = _null;
        }
      };

      _proto._onPointerEnd = function _onPointerEnd(position, id, event) {
        var pointerConstraint = this.pointerConstraint,
            target = pointerConstraint && pointerConstraint.target;

        if (target) {
          pointerConstraint.target = _null;
          target.emit('dragEnd', position, id, event);

          if (!target.active) {
            target.velocity.reset();
          }
        }
      };

      _proto.add = function add(object) {
        this.objects.push(object);
        return this;
      };

      _proto.remove = function remove(object) {
        var objects = this.objects,
            index = objects.indexOf(object);

        if (~index) {
          Utils.removeIndex(objects, index);
        }

        return this;
      };

      _proto.attach = function attach(object) {
        this.attachments.push(object);
        return this;
      };

      _proto.detach = function detach(object) {
        var attachments = this.attachments,
            index = attachments.indexOf(object);

        if (~index) {
          Utils.removeIndex(attachments, index);
        }

        return this;
      };

      _proto.use = function use(effect) {
        this.effects.push(effect);
        return this;
      };

      _proto.disuse = function disuse(effect) {
        var effects = this.effects,
            index = effects.indexOf(effect);

        if (~index) {
          Utils.removeIndex(effects, index);
        }

        return this;
      };

      _proto.toViewPosition = function toViewPosition(position) {
        var camera = this.camera;
        return camera ? camera.toViewPosition(position) : position;
      };

      _proto.toGlobalPosition = function toGlobalPosition(position) {
        var camera = this.camera;
        return camera ? camera.toGlobalPosition(position) : position;
      };

      _proto.getFocus = function getFocus(filter) {
        var pointerChecker = this.pointerChecker;

        if (pointerChecker) {
          var _pointer = this._pointer,
              bodies = this.objects.concat(this.attachments).filter(function (object) {
            return object.interactive && (!filter || filter(object));
          });

          for (var i = bodies.length; i--;) {
            var body = bodies[i];

            if (pointerChecker(_pointer, body)) {
              return body;
            }
          }
        }

        return _null;
      };

      _proto.drag = function drag(target) {
        var pointer = this.pointer;

        if (!pointer) {
          return;
        }

        var pointerConstraint = this.pointerConstraint;

        if (!pointerConstraint) {
          return;
        }

        var position = pointer.position;

        if (pointerConstraint.target) {
          pointerConstraint.target.emit('dragEnd');
        }

        pointerConstraint.target = target;

        if (target) {
          pointerConstraint.targetOffset.setVector(position).minusVector(target.position);
          target.emit('dragStart');
        }
      };

      _proto.update = function update(timeScale) {
        timeScale *= this.timeScale;
        this.emit('willUpdate', timeScale);
        var deferredEffects = this.effects.filter(function (effect) {
          if (effect.defer) {
            return true;
          } else {
            effect.update(timeScale);
          }
        });
        var pointerConstraint = this.pointerConstraint;

        if (pointerConstraint) {
          pointerConstraint.update(timeScale);
        }

        var collidableBodies = new Array();
        this.objects.filter(function (object) {
          if (object.update) {
            if (object.defer) {
              return true;
            } else {
              object.update(timeScale);
            }

            if (object.category && object.collisionFilter) {
              object.contact.clear();
              collidableBodies.push(object);
            }
          }
        }).forEach(function (object) {
          object.update(timeScale);
        });
        var target = pointerConstraint && pointerConstraint.target,
            _pointer = this._pointer;

        if (target && _pointer) {
          var position = _pointer.position;
          target.velocity.setVector(position).minusVector(this._pointerPosition);

          this._pointerPosition.setVector(position);
        }

        var collisionChecker = this.collisionChecker;

        if (collisionChecker) {
          Collision.check(collidableBodies, collisionChecker);
        }

        deferredEffects.forEach(function (effect) {
          effect.update(timeScale);
        });
        this.emit('didUpdate', timeScale);
      };

      _proto.render = function render(renderer) {
        renderer.fill(this.background);
        var camera = this.camera,
            context = renderer.context;

        if (camera) {
          camera.applyTo(context);
        }

        this.emit('willRender', renderer);
        this.objects.concat(this.attachments).forEach(function (renderable) {
          renderer.render(renderable);
        });
        this.emit('didRender', renderer);

        if (camera) {
          camera.restore(context);
        }
      };

      _createClass(Scene, [{
        key: "fps",
        set: function set(fps) {
          this.delay = 1000 / fps;
        },
        get: function get() {
          return 1000 / this.delay;
        }
      }, {
        key: "pointer",
        set: function set(pointer) {
          var pointerConstraint = this.pointerConstraint,
              _pointer = this._pointer;

          if (_pointer) {
            _pointer.off('click', this._onPointerClick);

            _pointer.off('start', this._onPointerStart);

            _pointer.off('end', this._onPointerEnd);

            if (!pointer && pointerConstraint && pointerConstraint.target) {
              pointerConstraint.target.emit('dragEnd');
            }
          }

          if (this._pointer = pointer) {
            pointer.on('click', this._onPointerClick);
            pointer.on('start', this._onPointerStart);
            pointer.on('end', this._onPointerEnd);
          }

          if (pointerConstraint) {
            pointerConstraint.origin = pointer && pointer.position;
          }
        },
        get: function get() {
          return this._pointer;
        }
      }]);

      return Scene;
    }(EventEmitter);
    Scene.defaults = {
      delay: Timer.defaults.delay,
      timeScale: 1,
      background: '#fff',
      clean: false,
      camera: _null,
      pointer: _null,
      collisionChecker: Collision.Checker.Smart,
      pointerChecker: Collision.Checker.Smart
    };

    var Inspector =
    /*#__PURE__*/
    function () {
      function Inspector(options) {
        _assign(this, Inspector.defaults, options);

        if (!this.paragraph) {
          this.paragraph = new Paragraph({
            position: Vector.of(-230, -150),
            lineHeight: 15,
            style: {
              font: '10px Consolas',
              textAlign: 'left',
              textBaseline: 'top',
              fillStyle: '#00f',
              strokeStyle: _null
            }
          });
        }
      }

      var _proto = Inspector.prototype;

      _proto.update = function update(engine) {
        var currentScene = engine.currentScene;
        this._objects = currentScene && currentScene.objects.concat(currentScene.attachments);
        this._camera = currentScene && currentScene.camera;
        this.paragraph.lines = this.callbacks.map(function (callback) {
          return callback(engine);
        });
      };

      _proto.render = function render(renderer) {
        var _objects = this._objects,
            _camera = this._camera,
            context = renderer.context;

        if (_objects) {
          if (_camera) {
            _camera.applyTo(context);
          }

          if (this.boundsStroke && _objects.length) {
            context.beginPath();

            _objects.forEach(function (object) {
              var bounds = object.bounds;

              if (bounds) {
                context.rect(bounds.left, bounds.top, bounds.width, bounds.height);
              }
            });

            context.strokeStyle = this.boundsStroke;
            context.lineWidth = this.boundsStrokeWidth;
            context.stroke();
          }

          if (this.velocityStroke) {
            var velocityStrokeScale = this.velocityStrokeScale;
            context.beginPath();

            _objects.forEach(function (object) {
              var velocity = object.velocity;

              if (velocity) {
                var position = object.position,
                    x = position.x,
                    y = position.y;
                context.moveTo(x, y);
                context.lineTo(x + velocity.x * velocityStrokeScale, y + velocity.y * velocityStrokeScale);
              }
            });

            context.strokeStyle = this.velocityStroke;
            context.lineWidth = this.velocityStrokeWidth;
            context.stroke();
          }

          if (_camera) {
            _camera.restore(context);
          }
        }

        renderer.render(this.paragraph);
      };

      return Inspector;
    }();
    Inspector.defaults = {
      callbacks: [function (engine) {
        var scene = engine.currentScene;
        return "FPS: " + (scene && scene.delay >= 0 ? (1000 / engine.timer.lastFrameDelay).toFixed(3) : 0);
      }, function (engine) {
        return "Frame Duration: " + engine.timer.lastFrameDuration;
      }, function (engine) {
        return "Objects: " + (engine.currentScene ? engine.currentScene.objects.length : 0);
      }, function (engine) {
        return "Effects: " + (engine.currentScene ? engine.currentScene.effects.length : 0);
      }, function (engine) {
        return "Attachments: " + (engine.currentScene ? engine.currentScene.attachments.length : 0);
      }, function (engine) {
        return "Pointer Position: " + engine.pointer.position;
      }],
      boundsStroke: _null,
      boundsStrokeWidth: 1,
      velocityStroke: _null,
      velocityStrokeWidth: 1,
      velocityStrokeScale: 1
    };

    var Engine =
    /*#__PURE__*/
    function (_EventEmitter) {
      _inheritsLoose(Engine, _EventEmitter);

      function Engine(options) {
        var _this;

        if (options === void 0) {
          options = Utils.Const.EMPTY_OBJECT;
        }

        _this = _EventEmitter.call(this) || this;
        _this.inspector = _null;
        _this.currentScene = _null;

        _assign(_assertThisInitialized(_this), Engine.defaults, options);

        if (!_this.timer) {
          _this.timer = new Timer();
        }

        if (!_this.renderer) {
          _this.renderer = new Renderer();
        }

        if (!_this.pointer) {
          _this.pointer = new Pointer();
        }

        _this.timer.on('tick', _this.tick = _this.tick.bind(_assertThisInitialized(_this)));

        var _assertThisInitialize = _assertThisInitialized(_this),
            renderer = _assertThisInitialize.renderer;

        _this.pointer.pretransform = renderer.toInnerPosition.bind(_this.renderer);
        renderer.on('resize', function () {
          if (_this.rerenderOnResize) {
            var _assertThisInitialize2 = _assertThisInitialized(_this),
                currentScene = _assertThisInitialize2.currentScene;

            if (currentScene) {
              var context = renderer.context;

              _this.emit('willRender', renderer);

              context.save();
              renderer.render(currentScene);
              context.restore();

              _this.emit('didRender', renderer);
            }
          }
        });
        return _this;
      }

      var _proto = Engine.prototype;

      _proto.createScene = function createScene(options) {
        return new Scene(_assign({
          pointer: this.pointer
        }, options));
      };

      _proto.enter = function enter(scene) {
        var currentScene = this.currentScene;

        if (currentScene) {
          currentScene.active = false;
          currentScene.emit('exit');
        }

        var pointer = this.pointer;
        this.currentScene = scene;

        if (scene) {
          var timer = this.timer;

          if (timer.delay !== scene.delay) {
            timer.reschedule(scene.delay);
          }

          pointer.transform = scene.toViewPosition;
          scene.active = true;
          scene.emit('enter');

          if (!timer.isRunning) {
            timer.start();
          }
        } else {
          pointer.transform = _null;
        }
      };

      _proto.tick = function tick(deltaTime) {
        if (deltaTime > this.maxDelay) {
          return;
        }

        var currentScene = this.currentScene,
            inspector = this.inspector,
            renderer = this.renderer,
            context = renderer.context;

        if (currentScene) {
          var timeScale = deltaTime / this.baseTime;
          this.emit('willUpdate', timeScale);
          var timer = this.timer;

          if (timer.delay !== currentScene.delay) {
            timer.reschedule(currentScene.delay);
          }

          currentScene.update(timeScale);
          this.emit('didUpdate', timeScale);
          this.emit('willRender', renderer);
          context.save();
          renderer.render(currentScene);
          context.restore();
          this.emit('didRender', renderer);
        }

        if (inspector) {
          inspector.update(this);
          renderer.render(inspector);
        }
      };

      return Engine;
    }(EventEmitter);
    Engine.defaults = {
      baseTime: 50,
      maxDelay: 1500,
      rerenderOnResize: true,
      restoration: true
    };

    exports.Body = Body;
    exports.Bounds = Bounds;
    exports.Camera = Camera;
    exports.Category = Category;
    exports.Circle = Circle;
    exports.Collision = Collision;
    exports.Constraint = Constraint;
    exports.Engine = Engine;
    exports.EventEmitter = EventEmitter;
    exports.Inspector = Inspector;
    exports.Layer = Layer;
    exports.Line = Line;
    exports.Paragraph = Paragraph;
    exports.Particles = Particles;
    exports.Pointer = Pointer;
    exports.Polygon = Polygon;
    exports.Pool = Pool;
    exports.Rectangle = Rectangle;
    exports.Renderer = Renderer;
    exports.Scene = Scene;
    exports.Shape = Shape;
    exports.Sizing = Sizing;
    exports.Sprite = Sprite;
    exports.Style = Style;
    exports.Text = Text;
    exports.Timer = Timer;
    exports.Timing = Timing;
    exports.Transition = Transition;
    exports.Utils = Utils;
    exports.Vector = Vector;
    exports.Vertices = Vertices;

    Object.defineProperty(exports, '__esModule', { value: true });

}));
