import { Timer } from "./Timer";
import { Renderer } from "../renderer/Renderer";
import { Scene, SceneOptions } from "./Scene";
import { Inspector } from "./Inspector";
import { Pointer } from "./Pointer";
import { EventEmitter } from "../common/EventEmitter";
export declare type EngineOptions = Partial<{
    timer: Timer;
    renderer: Renderer;
    pointer: Pointer;
    inspector: Inspector | null;
    baseTime: number;
    maxDelay: number;
    currentScene: Scene | null;
    rerenderOnResize: boolean;
    restoration: boolean;
}>;
export interface EngineEvents {
    willUpdate: number;
    didUpdate: number;
    willRender: Renderer;
    didRender: Renderer;
}
export declare class Engine extends EventEmitter<EngineEvents> implements Required<EngineOptions> {
    static defaults: EngineOptions;
    constructor(options?: Readonly<EngineOptions>);
    readonly timer: Timer;
    readonly renderer: Renderer;
    readonly pointer: Pointer;
    inspector: Inspector | null;
    baseTime: number;
    maxDelay: number;
    currentScene: Scene | null;
    rerenderOnResize: boolean;
    restoration: boolean;
    createScene(options?: SceneOptions): Scene;
    enter(scene: Scene | null): void;
    tick(deltaTime: number): void;
}
