import { Vector, VectorLike } from "../geometry/Vector";
import { EventEmitter } from "../common/EventEmitter";
import { BodyLike } from "../physics/Body";
import { Bounds } from "../geometry/Bounds";
export declare type PointerTransform = (position: Vector) => Vector;
export declare type PointerOptions = Partial<{
    target: EventTarget;
    pretransform: null | PointerTransform;
    transform: null | PointerTransform;
    isTouchMode: boolean;
    holdOnly: boolean;
    clickThreshold: number;
    radius: number;
}>;
export declare type PointerEventParameters = [Vector, number, Event];
export interface PointerEvents {
    start: PointerEventParameters;
    move: PointerEventParameters;
    end: PointerEventParameters;
    click: PointerEventParameters;
}
export declare class Pointer extends EventEmitter<PointerEvents> implements Required<PointerOptions>, BodyLike {
    static defaults: PointerOptions;
    constructor(options?: Readonly<PointerOptions>);
    readonly target: EventTarget;
    readonly isTouchMode: boolean;
    readonly isCircle = true;
    readonly position: Vector;
    readonly bounds: Bounds;
    readonly normals: [];
    readonly active: boolean;
    readonly isHolding: boolean;
    readonly startTimeStamps: Map<number, number>;
    readonly positions: Map<number, Vector>;
    private _startListener;
    private _moveListener;
    private _endListener;
    holdOnly: boolean;
    clickThreshold: number;
    pretransform: null | PointerTransform;
    transform: null | PointerTransform;
    radius: number;
    private _setPosition;
    private _start;
    private _move;
    private _end;
    getClosest(target: VectorLike): Vector;
    project(direction: Vector): {
        min: number;
        max: number;
    };
    destroy(): void;
}
