import { Renderable, Renderer } from "../renderer/Renderer";
import { EventEmitter } from "../common/EventEmitter";
import { RenderingStyle } from "../graph/Style";
import { Utils } from "../common/Utils";
import { Body } from "../physics/Body";
import { CollisionChecker } from "../physics/Collision";
import { Pointer } from "./Pointer";
import { Vector } from "../geometry/Vector";
import { Constraint } from "../physics/Constraint";
import { Camera } from "./Camera";
export interface SceneEffect {
    defer?: boolean;
    update(timeScale: number): void;
}
export declare type SceneObject = Body | (Renderable & Partial<SceneEffect>);
export declare type SceneOptions = Partial<{
    delay: number;
    fps: number;
    timeScale: number;
    background: RenderingStyle | null;
    clean: boolean;
    camera: Camera | null;
    pointer: Pointer | null;
    objects: SceneObject[];
    attachments: Renderable[];
    effects: SceneEffect[];
    collisionChecker: CollisionChecker | null;
    pointerChecker: CollisionChecker | null;
    pointerConstraint: Constraint | null;
}>;
export interface SceneEvents {
    enter: [];
    willUpdate: number;
    didUpdate: number;
    willRender: Renderer;
    didRender: Renderer;
    exit: [];
}
export declare class Scene extends EventEmitter<SceneEvents> implements Required<SceneOptions> {
    static defaults: SceneOptions;
    constructor(options?: Readonly<SceneOptions>);
    active: boolean;
    delay: number;
    timeScale: number;
    background: RenderingStyle | null;
    clean: boolean;
    camera: Camera | null;
    objects: SceneObject[];
    attachments: Renderable[];
    effects: SceneEffect[];
    collisionChecker: CollisionChecker | null;
    pointerChecker: CollisionChecker | null;
    pointerConstraint: Constraint | null;
    private _pointer;
    private _pointerPosition;
    fps: number;
    pointer: Pointer | null;
    private _onPointerClick;
    private _draggableFilter;
    private _onPointerStart;
    private _onPointerEnd;
    add(object: SceneObject): this;
    remove(object: SceneObject): this;
    attach(object: SceneObject): this;
    detach(object: SceneObject): this;
    use(effect: SceneEffect): this;
    disuse(effect: SceneEffect): this;
    toViewPosition(position: Vector): Vector;
    toGlobalPosition(position: Vector): Vector;
    getFocus(filter?: Utils.Callback<void, Body, any>): Body | null;
    drag(target: Body | null): void;
    update(timeScale: number): void;
    render(renderer: Renderer): void;
}
