import { EventEmitter } from "../common/EventEmitter";
import { Utils } from "../common/Utils";
export declare type ScheduleCallback<T extends any[] = any[]> = Utils.Callback<void, T, void>;
export declare type TimerOptions = Partial<{
    delay: number;
    allowRAF: boolean;
    fixDelay: boolean;
}>;
export interface TimerEvents {
    tick: number;
}
export declare class Timer extends EventEmitter<TimerEvents> implements Required<TimerOptions> {
    static RAFThreshold: number;
    static defaults: TimerOptions;
    constructor(options?: Readonly<TimerOptions>);
    readonly lastFrameDelay: number;
    readonly lastFrameDuration: number;
    readonly isRunning: boolean;
    delay: number;
    allowRAF: boolean;
    fixDelay: boolean;
    private _timer?;
    private _usedRAF?;
    private _lastTickTime;
    private _lastStopTime;
    private _timeoutId;
    private _timeoutCallbacks;
    private _intervalId;
    private _intervalCallbacks;
    private _requestTick;
    private _tick;
    start(): void;
    stop(): void;
    reschedule(delay: number): void;
    setTimeout<T extends any[] = any[]>(callback: ScheduleCallback<T>, timeout: number, ...args: T): number;
    clearTimeout(id: number): void;
    setInterval<T extends any[] = any[]>(callback: ScheduleCallback<T>, interval: number, ...args: T): number;
    clearInterval(id: number): void;
}
