export interface VectorLike {
    x: number;
    y: number;
}
export declare class Vector {
    static of(x: number, y: number): Vector;
    static from(vectorLike: VectorLike): Vector;
    static plus(vectors: VectorLike[]): Vector;
    static mean(vectors: VectorLike[]): Vector;
    static dot(vector1: VectorLike, vector2: VectorLike): number;
    static cross(vector1: VectorLike, vector2: VectorLike): number;
    static minus(vector1: VectorLike, vector2: VectorLike): Vector;
    static project(sourceVector: VectorLike, directionVector: Vector): number;
    static projectVector(sourceVector: VectorLike, directionVector: Vector): Vector;
    static distance(vector1: VectorLike, vector2: VectorLike): number;
    static distribute(vector0: VectorLike, vector1: VectorLike, vector2: VectorLike, k1: number, k2: number, scale?: number): void;
    static random(startAngle?: number, endAngle?: number): Vector;
    constructor(x?: number, y?: number);
    x: number;
    y: number;
    clone(): Vector;
    set(x: number, y: number): this;
    setVector(vector: VectorLike, scale?: number): this;
    reset(): this;
    getNorm(): number;
    setNorm(norm: number): this;
    plus(dx: number, dy: number): this;
    plusVector(vector: VectorLike, scale?: number): this;
    minus(dx: number, dy: number): this;
    minusVector(vector: VectorLike, scale?: number): this;
    scale(scaleX: number, scaleY?: number, origin?: VectorLike): this;
    shrink(shrinkX: number, shrinkY?: number, origin?: VectorLike): this;
    grow(delta: number): this;
    rotate(radian: number, origin?: VectorLike): this;
    getAngle(): number;
    turn(clockwise?: boolean): this;
    reverse(): this;
    normalize(): this;
    isZero(): boolean;
    toString(fractionDigits?: number): string;
}
