import { Renderable, RendererLike } from "../renderer/Renderer";
import { Vector } from "../geometry/Vector";
import { SceneObject } from "../core/Scene";
import { Pool } from "../common/Pool";
import { Utils } from "../common/Utils";
export declare type ParticleInitializer<T extends SceneObject = SceneObject, U = unknown> = (particle: T, life: number) => U;
export declare type ParticleCleaner<T extends SceneObject = SceneObject, U = unknown> = (particle: T, legacy: U) => void;
export declare type ParticlesOptions<T extends SceneObject = SceneObject, U = unknown> = Partial<{
    visible: boolean;
    active: boolean;
    offset: Vector;
    pool: Pool<T> | null;
    minLife: number;
    maxLife: number;
    life: number;
    items: T[];
    minCount: number;
    maxCount: number;
    count: number;
    threshold: number;
    initializer: ParticleInitializer<T, U> | null;
    cleaner: ParticleCleaner<T, U> | null;
    sequential: boolean;
    loop: boolean;
}>;
export declare class Particles<T extends SceneObject = SceneObject, U = unknown> implements Required<ParticlesOptions<T, U>>, Renderable {
    static defaults: ParticlesOptions;
    constructor(options?: Readonly<ParticlesOptions<T, U>>);
    visible: boolean;
    active: boolean;
    offset: Vector;
    pool: Pool<T> | null;
    minLife: number;
    maxLife: number;
    items: T[];
    minCount: number;
    maxCount: number;
    initializer: ParticleInitializer<T, U> | null;
    cleaner: ParticleCleaner<T, U> | null;
    sequential: boolean;
    loop: boolean;
    private _items;
    private readonly _deadlines;
    private readonly _legacies;
    life: number;
    count: number;
    private _spawn;
    spawn: Utils.ThrottleWrapper<() => void>;
    threshold: number;
    clear(): void;
    private _remove;
    update(timeScale: number): void;
    render(renderer: RendererLike): void;
}
