import { Body, BodyOptions } from "../physics/Body";
import { Renderable, RendererLike } from "../renderer/Renderer";
import { StrokeStyle, CommonStyle, FillStyle } from "./Style";
export declare type ShapeStyle = CommonStyle & StrokeStyle & FillStyle;
export declare type ShapeOptions = BodyOptions & Partial<{
    style: Partial<ShapeStyle>;
    visible: boolean;
    fillFirst: boolean;
    closePath: boolean;
    preferShadow: boolean;
    texture: Renderable | null;
    attachments: Renderable[];
}>;
export declare abstract class Shape extends Body implements Required<ShapeOptions>, Renderable {
    static defaults: ShapeOptions;
    static defaultStyle: ShapeStyle;
    static applyStyle(renderer: RendererLike, style: ShapeStyle): void;
    constructor(options?: Readonly<ShapeOptions>);
    style: ShapeStyle;
    visible: boolean;
    fillFirst: boolean;
    closePath: boolean;
    preferShadow: boolean;
    texture: Renderable | null;
    attachments: Renderable[];
    abstract path(context: CanvasRenderingContext2D): void;
    abstract updateBounds(): void;
    render(renderer: RendererLike): void;
}
