import { Renderable, RendererLike } from "../renderer/Renderer";
import { Vector } from "../geometry/Vector";
import { Utils } from "../common/Utils";
import { CommonStyle } from "./Style";
export declare type ImageLike = Exclude<CanvasImageSource, SVGImageElement>;
export declare type SpriteOptions = Partial<{
    visible: boolean;
    position: Vector;
    image: ImageLike;
    style: Partial<CommonStyle>;
    clipX: number;
    clipY: number;
    clipWidth: number;
    clipHeight: number;
    width: number;
    height: number;
    rotation: number;
}>;
export declare class Sprite implements Required<SpriteOptions>, Renderable {
    static defaults: SpriteOptions;
    static defaultStyle: CommonStyle;
    static of(src: string, options?: Readonly<SpriteOptions>): Sprite;
    constructor(options?: Readonly<SpriteOptions>);
    visible: boolean;
    position: Vector;
    image: ImageLike;
    style: CommonStyle;
    clipX: number;
    clipY: number;
    clipWidth: number;
    clipHeight: number;
    width: number;
    height: number;
    rotation: number;
    load(src: string, onSuccess?: Utils.Callback<GlobalEventHandlers, Event, void>, onFailure?: Utils.Callback<GlobalEventHandlers, [Event | string], void>): this;
    render(renderer: RendererLike): void;
}
