import { SizingFunction } from "./Sizing";
import { Vector, VectorLike } from "../geometry/Vector";
import { Utils } from "../common/Utils";
import { Bounds } from "../geometry/Bounds";
import { RenderingStyle } from "../graph/Style";
import { LayerOptions, Layer } from "../graph/Layer";
import { EventEmitter } from "../common/EventEmitter";
export interface Renderable {
    render(renderer: RendererLike): void;
}
export interface RendererLike {
    readonly canvas: HTMLCanvasElement;
    readonly context: CanvasRenderingContext2D;
    readonly ratio: number;
    readonly width: number;
    readonly height: number;
    readonly origin: Vector;
    readonly originX: number;
    readonly originY: number;
    fill(color: RenderingStyle | null): void;
}
export declare type RendererOptions = Partial<{
    canvas: HTMLCanvasElement;
    settings: CanvasRenderingContext2DSettings;
    width: number;
    height: number;
    margin: number;
    ratio: number;
    origin: Vector;
    parent: Element | null;
    align: boolean;
    sizing: SizingFunction;
    resizeEvents: string[];
    resizeDelay: number;
    restoration: boolean;
}>;
export interface RendererEvents {
    resize: [];
}
export declare class Renderer extends EventEmitter<RendererEvents> implements Required<RendererOptions>, RendererLike {
    static defaults: RendererOptions;
    constructor(options?: Readonly<RendererOptions>);
    readonly canvas: HTMLCanvasElement;
    readonly context: CanvasRenderingContext2D;
    readonly settings: CanvasRenderingContext2DSettings;
    readonly parent: Element | null;
    readonly resizeEvents: string[];
    readonly bounds: Bounds;
    readonly resizeListener: Utils.DebounceWrapper<() => void>;
    readonly origin: Vector;
    readonly originX: number;
    readonly originY: number;
    readonly width: number;
    readonly height: number;
    readonly margin: number;
    readonly ratio: number;
    align: boolean;
    sizing: SizingFunction;
    restoration: boolean;
    private _width;
    private _height;
    private _offsetX;
    private _offsetY;
    private _scale;
    resizeDelay: number;
    private _resize;
    private _resizeListener;
    resize(width: number, height: number, ratio?: number): this;
    fill(color: RenderingStyle | null): void;
    render(renderable: Renderable): void;
    toInnerPosition(position: VectorLike): Vector;
    toOuterPosition(position: VectorLike): Vector;
    createLayer(options?: LayerOptions): Layer;
}
